;
    Title "EDUC-8 Microcomputer - Keypad Module"

; The program material contained herein is supplied without representation
; or warranty of any kind. The author therefore assumes no responsibility and
; shall have no liability, consequential or otherwise, of any kind arising
; from the use of this program material or any part thereof.
; This software listing is also for private use only and not to be used for
; profit making purposes.
;
; Last Updated: 100119
;
; Copyright: A. Nixon 2019
;
; ----------------------------------------------------------------
; See KEY Module Circuit in help file to understand code operation
; ----------------------------------------------------------------
;
	errorlevel -302, -305 ; Turn off banking message, default destination

#include "p16f18345.inc"

	Constant TestMode	= 0	; set = 1 to enable test mode

; PIC16F18345 Configuration Bit Settings

; CONFIG1
; __config 0xD78C
 __CONFIG _CONFIG1, _FEXTOSC_OFF & _RSTOSC_HFINT32 & _CLKOUTEN_OFF & _CSWEN_OFF & _FCMEN_OFF
; CONFIG2
; __config 0xE731
 __CONFIG _CONFIG2, _MCLRE_OFF & _PWRTE_ON & _WDTE_OFF & _LPBOREN_OFF & _BOREN_OFF & _BORV_LOW & _PPS1WAY_OFF & _STVREN_OFF & _DEBUG_OFF
; CONFIG3
; __config 0x3
 __CONFIG _CONFIG3, _WRT_OFF & _LVP_OFF
; CONFIG4
; __config 0x3
 __CONFIG _CONFIG4, _CP_OFF & _CPD_OFF

;
; --------------------
; DEFINE GPR REGISTERS
; --------------------
;
	CBLOCK 0x20

flagReg
row
activeRow
tempA
tempB
keyBuff
clkCycles

	ENDC

;
; ----------------
; PORT DEFINITIONS
; ----------------
;
LED	equ 5h		; PORTA

ioDA	equ 4h		; PORTB
ioRI	equ 5h		; PORTB
ioRO	equ 6h		; PORTB
ioCLK	equ 7h		; PORTB

keyC1	equ 0h		; PORTC
keyC2	equ 1h		; PORTC
keyC3	equ 2h		; PORTC
keyC4	equ 3h		; PORTC
keyR1	equ 4h		; PORTC
keyR2	equ 5h		; PORTC
keyR3	equ 6h		; PORTC
keyR4	equ 7h		; PORTC
;
; -----------------------
; flagReg BIT DEFINITIONS
; -----------------------
;
clkHILO	equ 0h
rstHILO	equ 1h
keyDown	equ 2h
buffFull	equ 3h
ctrl	equ 4h



altCTRL	equ b'00010000';	; alternate ctrl flag
;
; ----------------------------------
; PROGRAM START - NO INTERRUPTS USED
; ----------------------------------
;
	org 0x0000

	BANKSEL TRISA
	movlw b'00001111'		; RA5 = OUT, all others unused = input - but have pullups enabled
	movwf TRISA
	movlw b'11000000'		; ioRO, ioCLK = IN, ioDAT, ioRI = OUT
	movwf TRISB
	movlw b'11111111'		; keypad columns, rows all inputs
	movwf TRISC

	BANKSEL PORTA
	movlw b'00000000'		; RA5=LED, others, unused inputs
	movwf PORTA
	movlw b'11110000'		; IO interface ioCLK, ioDAT = HI, - ioRO = HI (EDUC-8 sees not ready for data transfer)
	movwf PORTB
	movlw b'00001111'		; all key COL lines LO, all key row lines HI
	movwf PORTC
;
; -----------------------------------
; SET RA31, RB6 WEAK PULLUPS ENABLED
; -----------------------------------
;
	BANKSEL WPUA
	movlw b'00001111'		; unused float high
	movwf WPUA
	movlw b'11000000'		; ioCLK, ioRO float high
	movwf WPUB
	movlw b'11111111'		; all key lines float high
	movwf WPUC
;
; -------------------------
; SET A2D PINS = DIGITAL IO
; -------------------------
;
	BANKSEL ANSELA
	clrf ANSELA
	clrf ANSELB
	clrf ANSELC
;
; ----------------------
; DISABLE UNUSED MODULES
; ----------------------
;
	BANKSEL PMD0
	movlw b'11111111'
	movwf PMD0
	MOVWF PMD2
	movwf PMD3
	movwf PMD4
	movwf PMD5

	movlw b'11111110'		; TMR0 enabled
	movwf PMD1

	BANKSEL 0h
	movlw b'10000000'		; TMR0=ON - 8Bit
	movwf T0CON0
	movlw b'01001001'		; TMR0=Fosc/4, 1:512
	movwf T0CON1

	clrf PCLATH			; rom page 0 jumps
	clrf flagReg
	clrf row
	clrf activeRow
;
; --------------
; MAIN CODE LOOP
; --------------
;
MainLoop	movlb 0h			; RAM Bank 0
MLwait	btfsc PIR0,TMR0IF		; timer loop ~4mS
	goto genLoop
;
; when a key is pressed, the program sets the ioRI line LO to tell EDUC-8 that a key has been pressed and data is ready
; the EDUC-8 will respond by sending 8 clock pulses
; on the falling CLK edge, data is output to the ioDA line
; the EDUC-8 will capture this data on the rinsing edge
; once all 8 bits are send, the EDUC-8 will set the ioRO line LO briefly
; this signal is there to reset the keypad ready for a new keypress, ie set ioRI = HI
;
; -----------------------------------------
	btfsc flagReg,rstHILO		; testing for ioRO LO or HI signal - is normally HI on start - should float HI unconnected
	goto waitRSTHI

	btfsc PORTB,ioRO		; test for LO reset signal - if so, set ioRI = HI
	goto tstCLKHL

	bcf flagReg,buffFull		; key buffer empty
	bsf PORTB,ioRI		; EDUC-8 can see a HI signal, meaning the LED display is not ready for new data
	bsf flagReg,rstHILO		; wait for reset HI next
	goto tstCLKHL

waitRSTHI	btfsc PORTB,ioRO		; test for HI reset signal - if so, reset test flag
	bcf flagReg,rstHILO		; wait for reset LO next

; -----------------------------------------
tstCLKHL	btfsc flagReg,clkHILO		; testing for LO or HI CLK signal - is normally HI on start - should float HI unconnected
	goto waitCLKHI

	btfsc PORTB,ioCLK		; test for LO clock signal - if so, ready data for EDUC-8 to read on rising clock
	goto MLwait

	bcf STATUS,C		; shift in 0's
	rrf keyBuff
	btfsc STATUS,C
	goto keyBit1

	bcf PORTB,ioDA		; output 0
	bsf flagReg,clkHILO		; wait for CLK High next
	goto MLwait

keyBit1	bsf PORTB,ioDA		; output 1
	bsf flagReg,clkHILO		; wait for CLK High next
	goto MLwait

waitCLKHI	btfsc PORTB,ioCLK		; test for HI clock signal - if so, input 8 bit data on rising pulse
	bcf flagReg,clkHILO		; wait for LO CLK next
	goto MLwait

; -----------------------------------------
genLoop	bcf PIR0,TMR0IF		; reset tmr0 overflow flag

; -----------------------------------------
updtKeys	movf row,W
	andlw 0x07
	addwf PCL
	goto row1
	goto row2
	goto row3
	goto row4
	goto row4key
	goto MLwait			; in case data error
	goto MLwait			; in case data error
	goto MLwait			; in case data error

row1	movlw b'11101111'		; row 1 OUT LO - wait until next loop tom read - gives 16mS debounce
	goto endKey

row2	movlw d'0'			; keyBase value for row 1
	call setKeyVal		; yes
	movlw b'11011111'		; row 2 OUT LO
	goto endKey

row3	movlw d'4'			; keyBase value for row 2
	call setKeyVal		; yes
	movlw b'10111111'		; row 3 OUT LO
	goto endKey

row4	movlw d'8'			; keyBase value for row 3
	call setKeyVal		; yes
	movlw b'01111111'		; row 4 OUT LO
	goto endKey

row4key	movlw d'12'			; keyBase value for row 4
	call setKeyVal		; yes
	movlw 0xFF
	movwf row			; start scan again
	movlw b'11111111'
endKey
	BANKSEL TRISC
	movwf TRISC
	BANKSEL 0h
	incf row
	goto MLwait

;
; -----------------------------
; SUBROUTINE: PROCESS KEY PRESS
; -----------------------------
; W = KeyValue to start from
; produces key value to send to EDUC-8
setKeyVal	
	if TestMode == 1
	  return			; ignore keypress subroutine if in testmode
	endif

	movwf tempB
	movf PORTC,W		; read rows from active row
	andlw b'00001111'
	movwf tempA

	movf tempB,W
	btfss tempA,0
	goto newKeyVal		; ool + 1

	addlw 1h
	btfss tempA,1
	goto newKeyVal		; col + 2

	addlw 1h			; col + 3
	btfss tempA,2
	goto newKeyVal

	addlw 1h			; col + 4
	btfss tempA,3
	goto newKeyVal

; no key pressed for this column
; if active column matches the scan column value then reset keypress flag
	movf row,W
	xorwf activeRow,W
	btfss STATUS,Z
	return

	bcf flagReg,keyDown
	return

newKeyVal	btfsc flagReg,keyDown		; ignore keypress if key already down
	return

	movwf tempA
	bsf flagReg,keyDown		; key is pressed
	movf row,W
	movwf activeRow

	movf tempA,W
	xorlw d'15'			; if ctrl key, alternate CTRL flag
	btfss STATUS,Z
	goto notCtrl

	movlw altCTRL
	xorwf flagReg

	btfsc flagReg,ctrl
	goto LEDon

	bcf PORTA,LED
	return

LEDon	bsf PORTA,LED
	return

notCtrl	btfsc flagReg,buffFull		; see if key buffer still full
	return

	movf tempA,W
	btfsc flagReg,ctrl
	addlw 0x10
	movwf keyBuff

	bcf PORTB,ioRI		; alert EDUC-8 that key data is ready to capture - reset when EDUC-8 responds
	bsf flagReg,buffFull		; key buffer full
	return

	end


