;
    Title "EDUC-8 Microcomputer - LED Display Module"

; The program material contained herein is supplied without representation
; or warranty of any kind. The author therefore assumes no responsibility and
; shall have no liability, consequential or otherwise, of any kind arising
; from the use of this program material or any part thereof.
; This software listing is also for private use only and not to be used for
; profit making purposes.
;
; Last Updated: 100119
;
; Copyright: A. Nixon 2019
;
; ----------------------------------------------------------------
; See LED Module Circuit in help file to understand code operation
; ----------------------------------------------------------------
;
;

	errorlevel -302, -305 ; Turn off banking message, default destination

#include "p16f18345.inc"

; PIC16F18345 Configuration Bit Settings

; CONFIG1
; __config 0xD78C
 __CONFIG _CONFIG1, _FEXTOSC_OFF & _RSTOSC_HFINT32 & _CLKOUTEN_OFF & _CSWEN_OFF & _FCMEN_OFF
; CONFIG2
; __config 0xE731
 __CONFIG _CONFIG2, _MCLRE_OFF & _PWRTE_ON & _WDTE_OFF & _LPBOREN_OFF & _BOREN_OFF & _BORV_LOW & _PPS1WAY_OFF & _STVREN_OFF & _DEBUG_OFF
; CONFIG3
; __config 0x3
 __CONFIG _CONFIG3, _WRT_OFF & _LVP_OFF
; CONFIG4
; __config 0x3
 __CONFIG _CONFIG4, _CP_OFF & _CPD_OFF

;
; --------------------
; DEFINE GPR REGISTERS
; --------------------
;
	CBLOCK 0x20

flagReg
column
digU
digHT
tempA
delayH
waitH
waitL
DataIn
clkCycles

	ENDC

;
; ----------------
; PORT DEFINITIONS
; ----------------
;
AD	equ 0h		; A2D input
SLA	equ 1h		; PORTA
SLB	equ 3h		; PORTA
DC1	equ 2h		; PORTA
DC2	equ 4h		; PORTA
DC3	equ 5h		; PORTA

ioDA	equ 4h		; PORTB
ioRI	equ 5h		; PORTB
ioRO	equ 6h		; PORTB
ioCLK	equ 7h		; PORTB
;
; -----------------------
; flagReg BIT DEFINITIONS
; -----------------------
;
blankH	equ 0h
blankT	equ 1h
clkHILO	equ 2h
rstHILO	equ 3h

blankMask	equ b'11111100'

;
; ----------------------------------
; PROGRAM START - NO INTERRUPTS USED
; ----------------------------------
;
	org 0x0000

	BANKSEL TRISA
	movlw b'00001011'		; SLB SLA A2D = IN, DIGIT DRIVEs = OUT
	movwf TRISA
	movlw b'11010000'		; ioCLK, ioDA, ioRO = IN - ioRI = OUT
	movwf TRISB
	movlw b'00000000'		; LED SEGMENTS- RC7 unconnected output
	movwf TRISC

	BANKSEL PORTA
	movlw b'00110100'		; LED common's all high
	movwf PORTA
	movlw b'00000000'		; IO interface all lo - ioRO = LO = EDUC-8 sees ready for data transfer
	movwf PORTB
	movlw b'00000000'		; all segments low
	movwf PORTC
;
; -----------------------------------
; SET RA31, RB6 WEAK PULLUPS ENABLED
; -----------------------------------
;
	BANKSEL WPUA
	movlw b'00001010'		; SLA, SLB float high
	movwf WPUA
	movlw b'11010000'		; ioCLK, ioDA, ioRO float high
	movwf WPUB
;
; -----------
; SET A2D PIN
; -----------
;
	BANKSEL ANSELA
	movlw b'00000001'		; RA0 = analog, A2D = ON
	movwf ANSELA
	clrf ANSELB
	clrf ANSELC
;
; ----------------------
; DISABLE UNUSED MODULES
; ----------------------
;
	BANKSEL PMD0
	movlw b'11111111'
	movwf PMD0
	movwf PMD3
	movwf PMD4
	movwf PMD5

	movlw b'11111110'		; TMR0 enabled
	movwf PMD1
	movlw b'11011111'		; A2D enabled
	movwf PMD2

	BANKSEL ADCON0
	movlw b'00000001'		; RA0 for A2D, A2D=ON
	movwf ADCON0
	movlw b'00100000'		; Left Justified, Fosc/32, Refs - Vss, Vdd 
	movwf ADCON1

	BANKSEL 0h
	movlw b'10000000'		; TMR0=ON - 8Bit
	movwf T0CON0
	movlw b'01000111'		; TMR0=Fosc/4, 1:10
	movwf T0CON1

	clrf PCLATH			; rom page 0 jumps
	clrf flagReg
	clrf digHT
	clrf digU
	clrf DataIn
	clrf column
	movlw 8h			; receive 8 clock cycles for data input
	movwf clkCycles
	movlw d'2'			; LED ready delay -> need 2 - 20 seconds - delayH x secs delay (approx)
	movwf delayH
	clrf waitH
	goto MainLoop
;
; ----------------------------
; SUBROUTINE: 7 SEGMENT DECODE
; ----------------------------
;                   -GFEDCBA
seg7	addwf PCL
	retlw b'00111111'		; 0
	retlw b'00000110'		; 1
	retlw b'01011011'		; 2
	retlw b'01001111'		; 3
	retlw b'01100110'		; 4
	retlw b'01101101'		; 5
	retlw b'01111101'		; 6
	retlw b'00000111'		; 7
	retlw b'01111111'		; 8
	retlw b'01101111'		; 9
	retlw b'01110111'		; A
	retlw b'01111100'		; B
	retlw b'00111001'		; C
	retlw b'01011110'		; D
	retlw b'01111001'		; E
	retlw b'01110001'		; F
;
; --------------
; MAIN CODE LOOP
; --------------
;
MainLoop	movlb 0h			; RAM Bank 0
MLwait	btfsc PIR0,TMR0IF		; timer loop ~4mS
	goto genLoop

; -----------------------------------------
	btfsc flagReg,rstHILO		; testing for ioRO LO or HI signal - is normally HI on start - should float HI unconnected
	goto waitRSTHI

	btfsc PORTB,ioRO		; test for LO reset signal - if so, reset delay timer and set ioRI = HI
	goto tstCLKHL

	movf delayH,W
	movwf waitH
	clrf waitL
	movlw 8h			; receive another 8 clock cycles for data input
	movwf clkCycles
	bsf PORTB,ioRI		; EDUC-8 can see a HI signal, meaning the LED display is not ready for new data
	bsf flagReg,rstHILO		; wait for reset HI next
	goto tstCLKHL

waitRSTHI	btfsc PORTB,ioRO		; test for HI reset signal - if so, reset test flag
	bcf flagReg,rstHILO		; wait for reset LO next

; -----------------------------------------
tstCLKHL	btfsc flagReg,clkHILO		; testing for LO or HI CLK signal - is normally HI on start - should float HI unconnected
	goto waitCLKHI

	btfss PORTB,ioCLK		; test for LO clock signal - if so, flag input data on rising CLK edge
	bsf flagReg,clkHILO		; wait for CLK High next
	goto MLwait

waitCLKHI	btfss PORTB,ioCLK		; test for HI clock signal - if so, input 8 bit data on rising pulse
	goto MLwait

	bcf STATUS,C
	btfsc PORTB,ioDA
	bsf STATUS,C
	rrf DataIn
	bcf flagReg,clkHILO		; wait for LO CLK next
	decfsz clkCycles
	goto MLwait

	movlw 8h			; receive another 8 clock cycles for data input
	movwf clkCycles
	call nextDig		; format new data for display
	goto MLwait

; -----------------------------------------
genLoop	bcf PIR0,TMR0IF		; reset tmr0 overflow flag

; -----------------------------------------
; LED ready delay
	movf waitH,W		; if = 0, then delay already expired
	btfsc STATUS,Z
	goto updtDisp

	decfsz waitL		; decrement delay, if  = 0 then ioRI = LO
	goto updtDisp

	decf waitH
	btfsc STATUS,Z
	bcf PORTB,ioRI		; EDUC-8 can see a LO signal, meaning the LED display is ready for new data

; -----------------------------------------
; displays update every 4mS
updtDisp	movlw b'00110100'		; LED common's all high
	movwf PORTA
	movf column,W
	andlw 0x03
	addwf PCL
	goto digit1
	goto digit2
	goto digit3
	goto clrDig			; in case value error

digit1	swapf digHT,W
	andlw 0x0F
	call seg7
	movwf PORTC
	btfss flagReg,blankH		; blanked ??
	bcf PORTA,DC1		; digit 1 ON
	incf column
	goto MLwait

digit2	movf digHT,W
	andlw 0x0F
	call seg7
	movwf PORTC
	btfss flagReg,blankT		; blanked ??
	bcf PORTA,DC2		; digit 2 ON
	incf column
	BANKSEL ADCON0
	bsf ADCON0,GO		; start A2D conversion
	goto MainLoop

digit3	movf digU,W
	andlw 0x0F
	call seg7
	iorlw b'10000000'		; light DP on this digit
	movwf PORTC
	bcf PORTA,DC3		; digit 3 ON
clrDig	clrf column

; ---------------------------------
; update digit3 also reads analog and sets [display renew] delay (every 48mS)
;
	BANKSEL ADRESH
	swapf ADRESH,W		; need 2 - 20 approx from A2D result for delayH
	andlw b'00001110'
	BANKSEL 0h			; 00010 - 10100 - use 0010 - 1111 gives 2 - 15 seconds
	movwf tempA			; temp save new delay value

	movf delayH,W		; if delay has changed, update new delay and current timer
	xorwf tempA,W
	btfsc STATUS,Z
	goto MLwait

	movf tempA,W		; changed value read from A2D
	movwf delayH
	movwf waitH
	clrf waitL
	goto MLwait

; ---------------------------------
nextDig	movlw blankMask		; clear leading zero blanking flags
	andwf flagReg

	btfsc PORTA,SLA		; if = 0 then OCTAL format
	goto toBCDhex		; else = decimal format

	btfsc PORTA,SLB		; if = 0 then EDUC-8 octal, else 377 octal
	goto oct377

; OCT 737 -------------------------
	rrf DataIn,W		; oct737 - upper 3 bits to hundereds digit
	andlw b'01110000'
	movwf digHT
	rrf DataIn,W		; bits 43 to tens digit
	movwf tempA
	rrf tempA
	rrf tempA,W
	andlw b'00000011'
	iorwf digHT

	movf DataIn,W		; lower 3 bits
	andlw b'00000111'
	movwf digU
	return

; OCT 377 -------------------------
oct377	rrf DataIn,W		; upper 2 bits to hundereds digit
	movwf digHT
	rrf digHT,W
	andlw b'00110000'
	movwf digHT

	rrf DataIn,W		; bits 543 to bits 210
	movwf tempA			; then to tens digit
	rrf tempA
	rrf tempA,W
	andlw b'00000111'
	iorwf digHT

	movf DataIn,W		; bits 210 to digU
	andlw b'00000111'
	movwf digU
	goto flagsLZB		; set leading zero blanking flags

; BCD or HEX -----------------
toBCDhex	btfsc PORTA,SLB		; if = 0 then EDUC-8 octal, else 377 octal
	goto doBCD

; DECIMAL HEX ----------------
	swapf DataIn,W
	andlw 0x0F
	movwf digHT			; 1st digit not used
	bsf flagReg,blankH
	movf DataIn,W
	andlw 0x0F
	movwf digU
	return

; DECIMAL TO BCD -------------
doBCD	clrf digHT			; decimal has leading zero blanking
	decf digHT			; preset 'tens' to -1
	movf DataIn,W
div10	movwf digU
	incf digHT,F		; bump 'tens', 0x00..0x25
	movlw d'6'			; using "packed bcd" format
	addwf digHT,W		; bcd "digit carry"?
	skpndc			; no, skip, else
	movwf digHT			; fix 'tens'
	movlw d'10'			; ones = ones - 10
	subwf digU,W		; borrow?
	bc div10			; no, branch, else done

flagsLZB	movf digHT,W		; if H = 0 then H digit blank
	andlw 0xF0
	btfss STATUS,Z
	return

	bsf flagReg,blankH
	movf digHT,W		; if T = 0 then T digit blank
	andlw 0x0F
	btfsc STATUS,Z
	bsf flagReg,blankT
	return			; back to main loop

	end