;
    Title "EDUC-8 Microcomputer"

; The program material contained herein is supplied without representation
; or warranty of any kind. The author therefore assumes no responsibility and
; shall have no liability, consequential or otherwise, of any kind arising
; from the use of this program material or any part thereof.
; This software listing is for private use only and is not to be used for
; profit making purposes.
;
; Last Updated: 310718
;
; Copyright: A. Nixon 2018
;
#include P18F47K40.inc

	Constant TestMode	= 0	; 1 = enable test mode
;
; CRYSTAL SPEED = 20MHz
;

;*******************************************************************************
;
; TODO Step #2 - Configuration Word Setup
;
;*******************************************************************************

; TODO INSERT CONFIG HERE

; PIC18F47K40 Configuration Bit Settings

; CONFIG1L
  CONFIG  FEXTOSC = OFF        ; External Oscillator is disabled
  CONFIG  RSTOSC = HFINTOSC_64MHZ   ; Power-up default value for COSC bits (Internal Oscillator mode 64MHz 1:1)

; CONFIG1H
  CONFIG  CLKOUTEN = OFF        ; Clock Out Enable bit (CLKOUT function is disabled)
  CONFIG  CSWEN = ON            ; Clock Switch Enable bit (The NOSC and NDIV bits can be changed by user software)
  CONFIG  FCMEN = OFF           ; Fail-Safe Clock Monitor Enable bit (Fail-Safe Clock Monitor disabled)

; CONFIG2L
  CONFIG  MCLRE = EXTMCLR       ; Master Clear Enable bit (If LVP = 0, MCLR pin is MCLR; If LVP = 1, RE3 pin function is MCLR )
  CONFIG  PWRTE = ON            ; Power-up Timer Enable bit (Power up timer enabled)
  CONFIG  LPBOREN = OFF         ; Low-power BOR enable bit (ULPBOR disabled)
  CONFIG  BOREN = OFF           ; Brown-out Reset Enable bits (Brown-out Reset disabled)

 ; CONFIG2H
  CONFIG  BORV = VBOR_285       ; Brown Out Reset Voltage selection bits (Brown-out Reset Voltage (VBOR) set to 2.45)
  CONFIG  ZCD = OFF             ; ZCD Disable bit (ZCD disabled. ZCD can be enabled by setting the ZCDSEN bit of ZCDCON)
  CONFIG  PPS1WAY = OFF         ; PPSLOCK bit One-Way Set Enable bit (PPSLOCK bit can be set and cleared repeatedly (subject to the unlock sequence))
  CONFIG  STVREN = OFF          ; Stack Full/Underflow Reset Enable bit (Stack full/underflow will not cause Reset)
  CONFIG  DEBUG = OFF           ; Debugger Enable bit (Background debugger disabled)
  CONFIG  XINST = OFF           ; Extended Instruction Set Enable bit (Extended Instruction Set and Indexed Addressing Mode disabled)

; CONFIG3L
  CONFIG  WDTCPS = WDTCPS_31    ; WDT Period Select bits (Divider ratio 1:65536; software control of WDTPS)
  CONFIG  WDTE = OFF            ; WDT operating mode (WDT Disabled)

; CONFIG3H
  CONFIG  WDTCWS = WDTCWS_7     ; WDT Window Select bits (window always open (100%); software control; keyed access not required)
  CONFIG  WDTCCS = SC           ; WDT input clock selector (Software Control)

; CONFIG4L
  CONFIG  WRT0 = OFF            ; Write Protection Block 0 (Block 0 (000800-003FFFh) not write-protected)
  CONFIG  WRT1 = OFF            ; Write Protection Block 1 (Block 1 (004000-007FFFh) not write-protected)
  CONFIG  WRT2 = OFF            ; Write Protection Block 2 (Block 2 (008000-00BFFFh) not write-protected)
  CONFIG  WRT3 = OFF            ; Write Protection Block 3 (Block 3 (00C000-00FFFFh) not write-protected)
  CONFIG  WRT4 = OFF            ; Write Protection Block 3 (Block 4 (010000-013FFFh) not write-protected)
  CONFIG  WRT5 = OFF            ; Write Protection Block 3 (Block 5 (014000-017FFFh) not write-protected)
  CONFIG  WRT6 = OFF            ; Write Protection Block 3 (Block 6 (018000-01BFFFh) not write-protected)
  CONFIG  WRT7 = OFF            ; Write Protection Block 3 (Block 7 (01C000-01FFFFh) not write-protected)

; CONFIG4H
  CONFIG  WRTC = OFF            ; Configuration Register Write Protection bit (Configuration registers (300000-30000Bh) is not write-protected)
  CONFIG  WRTB = OFF            ; Boot Block Write Protection bit (Boot Block (000000-0007FFh) not write-protected)
  CONFIG  WRTD = OFF            ; Data EEPROM Write Protection bit (Data EEPROM not write-protected)
  CONFIG  SCANE = OFF           ; Scanner Enable bit (Scanner module is NOT available for use, SCANMD bit is ignored)
  CONFIG  LVP = OFF             ; Low Voltage Programming Enable bit (HV on MCLR/VPP must be used for programming)

; CONFIG5L
  CONFIG  CP = OFF              ; UserNVM Program Memory Code Protection bit (UserNVM code protection disabled)
  CONFIG  CPD = OFF             ; DataNVM Memory Code Protection bit (DataNVM code protection disabled)

; CONFIG5H

; CONFIG6L
  CONFIG  EBTR0 = OFF	        ; Table Read Protection Block 0 (Block 0 (000800-003FFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR1 = OFF           ; Table Read Protection Block 1 (Block 1 (004000-007FFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR2 = OFF           ; Table Read Protection Block 2 (Block 2 (008000-00BFFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR3 = OFF           ; Table Read Protection Block 3 (Block 3 (00C000-00FFFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR4 = OFF           ; Table Read Protection Block 4 (Block 4 (010000-013FFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR5 = OFF           ; Table Read Protection Block 5 (Block 5 (014000-017FFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR6 = OFF           ; Table Read Protection Block 6 (Block 6 (018000-01BFFFh) not protected from table reads executed in other blocks)
  CONFIG  EBTR7 = OFF           ; Table Read Protection Block 7 (Block 7 (01C000-01FFFFh) not protected from table reads executed in other blocks)

; CONFIG6H
  CONFIG  EBTRB = OFF           ; Boot Block Table Read Protection bit (Boot Block (000000-0007FFh) not protected from table reads executed in other blocks)

; *****************************************************************************************************************************************

; RAM definitions
XFERbuff		equ 1h		; RAM Block 1 - Boot Write - Program ROM buffer - needs 128 bytes, other have 256 byte
RAMbuff		equ 2h		; RAM Block 2 - Buffer for 64 bytes of RAM
MEMbuff		equ 3h		; RAM Block 3 - needs 256 bytes, or WIF needs 2048 bytes 1K * 2, 8 pages

; INTERNAL EEPROM MEMORY
; 0 = operational flags
IEEfgACT		equ 0x01		; storage flag - program active - 1st byte of program name
IEEfgZPM		equ 0x02		; storage flag - ZPM
IEEfgT13		equ 0x04		; storage flag - T1/T13
IEEfgWIF		equ 0x08		; storage flag - Normal/WIF

EDUCnames		equ 0x010		; Storage for 8 EDUC Program Names
WIFnames		equ 0x0D0		; Storage for 3 WIF Program Names
progNameSize	equ d'24'		; program name character size (23) + 1 1st byte for flags info

IntEEPwif		equ 0x3E0		; 32 bytes for WIF EEPROM storage

; Program INFO byte
pibValid		equ 0h		; Status,     Empty         = 1, valid   = 0
pibZPM		equ 1h		; ZPM mode,   NORM          = 1, ZPM     = 0
pibSED		equ 2h		; SED mode,   NORM          = 1, SED     = 0
pibT1T13		equ 3h		; Reset,      T13 (AO)      = 1, T1 (BO) = 0
prgNotValidMask	equ 0x01		; not valid set

SwichScanRate	equ d'13'		; 13 * 0.8mS ~ 10mS switch scan rate
initCycleIndex	equ 0xFC		; initial cycle jump table index (-4)
wiRAMmax		equ 0x3F		; max RAM size - 1     64 bytes
wiRAMPAGEmax	equ 0x1F		; max RAM Page size    32 bytes
wiEEPmax		equ 0x1F		; max EEP size         32 bytes
wiRomHI		equ 0x03		; max ROM high (0x3FF)
WIdevSelect		equ 0x20		; ROM Word L - xxdx xxxx
maskTGT                 equ 0x3F		; iCodeL TGT mask

iimNone		equ d'0'		; indexes set for GOTO jump table
iimDeposit		equ d'4'
iimExamine		equ d'8'
iimFetch		equ d'12'
iimDefer		equ d'16'
iimExecute		equ d'20'

kyfNone		equ d'0'
kyfLoadAddress	equ d'1'
kyfDeposit		equ d'2'
kyfExamine		equ d'3'

SlowMinH		equ 0x0A		; minimum slow run rate
SlowMinL		equ 0x2C

swScanRate		equ d'10'		; 10 display refreshes per switch check

;
; --------------------
; ACCESS GPR REGISTERS
; --------------------
;
	CBLOCK 0x00
PCregL
MAreg
MBreg
ACreg
insLEDs
rfdeLEDs
TempA
TempB
TempC
TempD
DelA
DelB
DelC
LEDupdate
SWscan
ScanDigit
genBits1
genBits2
execIndex
binSwitch
ctrSWnew
ctrSWold
ctrSWuse
INSreg
ROLregA
ROLregB
ROLregC
execFunction
cycleCountH
cycleCountL
SlowCalH
SlowCalL
SwRdDelay
fgsExecDefer
PCregH
genBitsWIF
wifADDR
wiCodeH
wiCodeL
stack0H
stack0L
stack1H
stack1L
stack2H
stack2L
wifTblMain
wifTbl013
wifTbl1034
MainDelay
LEDupCnt
EEaddrH
EEaddrL
EEcnt
EEdat
errorFlags
tempPC
tempMA
tempMB
tempAC
tempRFDE
tempINS
ROMaddrH
ROMaddrL
valTIMER
delTIMER
statusWIF
MULHreg
MULLreg
intEEPaddr
				; (0x5F is highest)
	ENDC
;
; PC MAIN LOOP COMMANDS
;
rzMainCmdCon	equ 0xA0		; is EDUC-8 connected to PC
rzMainCmdGoProg	equ 0xA1		; PC request to enter boot loader
rzMainCmdMemWT	equ 0xA2		; set memory to a value
rzMainCmdMemRD	equ 0xA3		; read memory
rzMainCmdMemEr	equ 0xA4		; erase memory
rzMainCmdOpts	equ 0xA5		; write internal EEPROM, READ current
rzMainCmdName	equ 0xA6		; read/write program name(s)
rzMainCmdWTprg	equ 0xA7		; write program to EXT EEP
rzMainCmdRDprg	equ 0xA8		; read program from EXT EEP
rzMainCmdDLprg	equ 0xA9		; erase program(s)
rzMainCmdBoot2	equ 0xAA		; start boot 2 for reprogramming boot code

rzMainCmdTest	equ 0xAF		; my switches test
;
; PC BOOT COMMANDS
;

rzBootCmdAddr	equ 00h
rzBootCmdInit	equ 01h
rzBootCmdFill	equ 02h
rzBootCmdProg	equ 03h
rzBootCmdEEread	equ 04h
rzBootCmdEEwrit	equ 05h

rzBootCmdExit	equ 07h
rzBootCmdNoProg	equ 0xF3
;
; PC External EEPROM COMMANDS
;
pcEEPaddr		equ 0x50
pcEEPdata		equ 0x51
pcEEPread		equ 0x52
pcEEPwrite		equ 0x53
pcEEPinit		equ 0x54
pcEEPend		equ 0x56
pcEEPerror		equ 0x58
;
; --------------------
; genBits1 Definitions
; --------------------
;
fgErase		equ 0h
fgSlowFast		equ 1h
fgSinCon		equ 2h
fgDeferSet		equ 3h
fgHALT		equ 4h
fgSHdata		equ 5h
fgSwPanFunc		equ 6h
fgNoPCPLusOne	equ 7h
;
; --------------------
; genBits2 Definitions
; --------------------
;
n_ack		equ 0h		; eeprom ACK flag
EEerror		equ 1h		; eeprom write error
InitSW		equ 2h		; initial switch state
cyclesWait		equ 3h		; stop run after function completes
doKeyScan		equ 4h
fgTemp		equ 5h
fgTimer		equ 6h

ToggleSWscan	equ b'00010000'	; toggle doKeyScan bit
;
; ----------------------
; genBitsWIF Definitions
; ----------------------
;
wifSkip		equ 0h
wifNewINS		equ 1h
wifDev		equ 2h
wifSerial		equ 3h
wifSerRcvd		equ 4h
wifPCmod		equ 5h


;
; ---------------------
; statusWIF Definitions
; ---------------------
;
wifCarry		equ 0h
wifZero		equ 1h
wifDC		equ 2h
wifRP		equ 3h

;
; ----------------------
; Error Flag Definitions
; ----------------------
;
efgERRdisplay	equ 0h
efgInitOff		equ 1h
efgOK		equ 2h
efgError01		equ 3h
efgError71		equ 4h
efgOnOff		equ 5h
efgExit		equ 6h

;	
; -----------------------------------
; INTERNAL EEPR0M Addr 0 = USER FLAGS
; -----------------------------------
;
EEflgWIF		equ 0h		; Normal WIF mode

EEflgZPM		equ 2h		; Zero Page Mode
EEflgSED		equ 3h		; Allow slow ex/dp
EEflgSEXL		equ 4h		; slow execution flag L
EEflgSEXH		equ 5h		; slow execution flag H
EEflgT13		equ 6h		; T1 T13 Reset
EEflgRst		equ 7h		; clear registers on reset
maskSlowExec	equ b'11001111'	; slow exec mode = 125mS
;
; -------------------------
; CTRL switches definitions
; -------------------------
;
ctrsSF		equ 0h
ctrsSC		equ 1h
ctrsHL		equ 2h
ctrsRN		equ 3h
ctrsEX		equ 4h
ctrsDP		equ 5h
ctrsLA		equ 6h

maskRNHLSCSF	equ b'00001111'
maskLADPEX		equ b'01110000'
;
; ------------------
; execution steering
; ------------------
;
funcDeposit		equ d'4'
funcExamine		equ d'8'
funcFetch		equ d'12'
funcDefer		equ d'16'
funcExecute		equ d'20'
;
; -----------------------
; rfdeLEDs Bit Definitions
; -----------------------
;
Run		equ 7h
Fetch		equ 6h
Defer		equ 5h
Execute		equ 4h
WIF		equ 3h
SlowExamDep		equ 2h
ResetT1T13		equ 1h
PageZero		equ 0h

maskRFDEoff		equ 0x0F
maskSetRun		equ 0x80
maskSetFetch	equ 0x40
maskSetDefer	equ 0x20
maskSetExecute	equ 0x10
maskSetRunFetch	equ 0xC0
maskSetRunDefer	equ 0xA0
maskSetRunExec	equ 0x90
maskClrFlags	equ 0xF0

; WIF
winRun		equ 7h
winRead		equ 6h
winExec		equ 5h
winWrite		equ 4h
wiWIF		equ 3h
wiAOBO		equ 2h
wiHR		equ 1h
wiLR		equ 0h
;
; -----------------------
; insLEDs Bit Definitions
; -----------------------
;
oprAND		equ 0x80		; Board A type
oprTAD		equ 0x40
oprISZ		equ 0x20
oprDCA		equ 0x10
oprOPR		equ 0x08
oprIOT		equ 0x04
oprJMP		equ 0x02
oprJMS		equ 0x01

wstI		equ 7h
wstAC		equ 6h
wstREG		equ 5h
wstPC		equ 4h
winREG		equ 3h
winVAL		equ 2h
winGEN		equ 1h
winIO		equ 0h

insVAL		equ b'00000100'
insGEN		equ b'00000010'

iSRC_A		equ 0x00
iSRC_R		equ 0x01
iSRC_I		equ 0x02
iSRC_P		equ 0x03
iSRC_Mask		equ b'00000011'

iTGT_A		equ 0x00
iTGT_R		equ 0x40
iTGT_I		equ 0x80
iTGT_P		equ 0xC0
iTGT_Mask		equ b'11000000'

picNOP		equ d'0'		; 00
picADDVA		equ d'1'
picANDVA		equ d'2'
picIORVA		equ d'3'
picMOVVA		equ d'4'
picSUBVA		equ d'5'
picXORVA		equ d'6'
picSKIPA		equ d'7'
picRDSWA		equ d'8'
picDAAC		equ d'9'
picMULTA		equ d'10'
picRANDA		equ d'11'
;		equ d'12'
;		equ d'13'
;		equ d'14'
;		equ d'15'

picADDR		equ d'16'		; 01
picANDR		equ d'17'
picCLRR		equ d'18'
picDECR		equ d'19'
picDECSZ		equ d'20'
picINCR		equ d'21'
picINCSZ		equ d'22'
picIORR		equ d'23'
picMOVR		equ d'24'
picROTL		equ d'25'
picROTR		equ d'26'
picSUBR		equ d'27'
picXORR		equ d'28'
picEXCHR		equ d'29'
picCOMPR		equ d'30'
;		equ d'31'

picGOSUB		equ d'32'		; 10
picGOTO		equ d'33'
picRETURN		equ d'34'
picRETVAL		equ d'35'
picHALT		equ d'36'
picSTSKIP		equ d'37'
picSTFLAG		equ d'38'
picSETB		equ d'39'
picCLRB		equ d'40'
picTOGB		equ d'41'
picSKIPBS		equ d'42'
picSKIPBC		equ d'43'
picROMaddr		equ d'44'




picSKIPIF		equ d'48'		; 11
picREADID		equ d'49'
picRESETI		equ d'50'
picRDRSTI		equ d'51'
picSKIPOF		equ d'52'
picWRITOD		equ d'53'
picRESETO		equ d'54'
picWTRSTO		equ d'55'
picSEROUT		equ d'56'
picSKIPNS		equ d'57'
picSERACC		equ d'58'
picROMACC		equ d'59'
picSKIPTS		equ d'60'
picTIMER		equ d'61'
picEEPROM		equ d'62'
;		equ d'63'

insNOP		equ 0xA400		; instruction value for a NOP - used for substituing an invalid instruction - see isValidIns
;
; -------------------
; INPUT/OUTPUT SELECT
; -------------------
;
DATOUT1		equ 2h		; RB2
OUT1RI		equ 3h		; RB3
CLKA		equ 4h		; RB4
CLKB		equ 5h		; RB5

IN1DA		equ 0h		; RC0
IN1RI		equ 1h		; RC1
IN0DA		equ 2h		; RC2
IN0RI		equ 3h		; RC3
DATOUT0		equ 4h		; RC4
OUT0RI		equ 5h		; RC5


#Define	skipIN0DAlo		btfsc PORTC,IN0DA
#Define	skipIN1DAlo		btfsc PORTC,IN1DA
#Define	skipIN0RIlo		btfsc PORTC,IN0RI
#Define	skipIN1RIlo		btfsc PORTC,IN1RI
#Define	skipIN0RIhi		btfss PORTC,IN0RI
#Define	skipIN1RIhi		btfss PORTC,IN1RI
#Define	skipOUT0RIlo	btfsc PORTC,OUT0RI
#Define	skipOUT1RIlo	btfsc PORTB,OUT1RI
#Define	skipOUT0RIhi	btfss PORTC,OUT0RI
#Define	skipOUT1RIhi	btfss PORTB,OUT1RI

#Define	clrIOclk		bcf PORTB,CLKA
#Define	setIOclk		bsf PORTB,CLKA
#Define	clrIOrst		bcf PORTB,CLKB
#Define	setIOrst		bsf PORTB,CLKB

#Define	clrDataOut0		bcf PORTC,DATOUT0
#Define	clrDataOut1		bcf PORTB,DATOUT1
#Define	setDataOut0		bsf PORTC,DATOUT0
#Define	setDataOut1		bsf PORTB,DATOUT1

selIOin0	MACRO		; 11
	movf PORTB,W
	andlw b'11111100'
	iorlw b'00000011'
	movwf PORTB
	ENDM
selIOin1	MACRO		; 10
	movf PORTB,W
	andlw b'11111100'
	iorlw b'00000010'
	movwf PORTB
	ENDM
selIOout0	MACRO		; 01
	movf PORTB,W
	andlw b'11111100'
	iorlw b'00000001'
	movwf PORTB
	ENDM
selIOout1	MACRO		; 00
	movf PORTB,W
	andlw b'11111100'
	movwf PORTB
	ENDM

SetWriteEE	MACRO
	movlw b'00000100'		; Setup Data EEPROM Access, enable writes
	movwf NVMCON1
	ENDM
SetReadROM	MACRO
	movlw b'10000000'		; Setup Data ROM Access, reading
	movwf NVMCON1
	ENDM
;
; ---------------
; I2C DEFINITIONS
; ---------------
;
#Define	SkipIfSDA_Low	btfsc PORTA,7
#Define	SkipIfSDA_Hi	btfss PORTA,7
#Define	SCL_Hi		bsf PORTA,6
#Define	SCL_Lo		bcf PORTA,6
#Define	SDA_Hi		bsf TRISA,7
#Define	SDA_In		bsf TRISA,7
SDA_Lo	MACRO
	bcf LATA,7
	bcf TRISA,7
	ENDM
SDA_Out	MACRO
	bcf LATA,7
	bcf TRISA,7
	ENDM
EEPread	equ b'10100001'		; I2C address 0 read
EEPwrite	equ b'10100000'		; I2C address 0 write
;
; ----------
; CODE START
; ----------
;
	org 0x0000

	if TestMode == 1		; PC program will check Adress 0x000000 to see if it = 0, if so then report TestMode error
	  nop
	endif

	BANKSEL OSCFRQ
	movlw b'00000110'		; select 32MHz internal clock frequency
	movwf OSCFRQ,BANKED
; RA0 OL LK0 
; RA1 OL LK1
; RA2 OL LK2
; RA3 OL LK3
; RA4 OL LK4
; RA5 OL LK5
; RA6 OH I2C SCL 
; RA7 IN I2C SDA

; RB0 OL SELA
; RB1 OL SELB
; RB2 OH OUT1DA
; RB3 IN OUT1RI
; RB4 OH CLKA
; RB5 OH CLKB
; RB6 IN DEP0/ISCP
; RB7 IN LAD0/ICSP

; RC0 IN IN1DA
; RC1 IN IN1RI
; RC2 IN IN0DA
; RC3 IN IN0RI
; RC4 OH OUT0DA
; RC5 IN OUT0RI
; RC6 OH TX PICout
; RC7 IN RX PICin

; RD0 OL LA0
; RD1 OL LA1
; RD2 OL LA2
; RD3 OL LA3
; RD4 OL LA4
; RD5 OL LA5
; RD6 OL LA6
; RD7 OL LA7

; RE0 IN CTRL SW
; RE1 IN DIG SW
; RE2 IN EXT SW
	movlw b'01000000'
	movwf LATA
	movlw b'00000000'
	movwf LATB
	movlw b'01000000'
	movwf LATC
	movlw b'00000000'
	movwf LATD
	movlw b'00000000'
	movwf LATE

	movlw b'10000000'
	movwf TRISA
	movlw b'11001000'
	movwf TRISB
	movlw b'10101111'
	movwf TRISC
	movlw b'00000000'
	movwf TRISD
	movlw b'11111111'
	movwf TRISE
;
; DIGIT IO ON ALL PINS
;
	BANKSEL ANSELA
	clrf ANSELA,BANKED		; digital IO on portA
	clrf ANSELB,BANKED		; digital IO on portB
	clrf ANSELC,BANKED		; digital IO on portC
	clrf ANSELD,BANKED		; digital IO on portD
	clrf ANSELE,BANKED		; digital IO on portE
;
; weak pullups enabled for PORTD and PORTE 0 1 pins
;
	BANKSEL WPUD
	movlw 0xFF
	movwf WPUD
	movlw b'00000011'
	movwf WPUE
;
; -------------------------------------------------------
; PORTA,6 = OPEN DRAIN FOR SDA EEPROM PIN - RE01 AS WELL
; -------------------------------------------------------
;
; **** ONLY FOR BOARD 2 ****
	BANKSEL ODCONA
	movlw b'10000000'		; RA7
	movwf ODCONA,BANKED
	movlw b'00000011'		; RE1 RE0
	movwf ODCONE,BANKED
;
; ----------------------
; DISABLE UNUSED MODULES
; ----------------------
;
	BANKSEL PMD0
	movlw b'01111011'		; leave system clock on for EUSART - uses Fosc, HLVD = OFf
	movwf PMD0,BANKED
	movlw b'11111110'		; tmr0 = ON, tmr1 = OFF
	movwf PMD1,BANKED
	movlw b'11111111'
	movwf PMD2,BANKED
	movlw b'11111111'
	movwf PMD3,BANKED
	movlw b'10111111'		; EUSART = ON
	movwf PMD4,BANKED
	movlw b'11111111'
	movwf PMD5,BANKED
;
; NO SLEW RATE
;
	BANKSEL SLRCONA
	clrf SLRCONA,BANKED
	clrf SLRCONB,BANKED
	clrf SLRCONC,BANKED
	clrf SLRCOND,BANKED
	clrf SLRCONE,BANKED
;
; -------------------------------------------
; SET BAUD RATE TO COMMUNICATE WITH PC @12MHz
; -------------------------------------------
; Boot Baud Rate = 19200, No Parity, 1 Stop Bit
;
	movlw 0x01			; 19200 baud @32MHz
	movwf SP1BRGH
	movlw 0xA0
	movwf SP1BRGL
	movlw b'00100100'		; ENABLED, SYNC = 0, BRGH = 1
	movwf TX1STA
	movlw b'00001000'		; BRG16 = 1
	movwf BAUD1CON
	movlw b'10010000'		; enable Async Reception
	movwf RC1STA

; Bank to PPSLOCK register
	BANKSEL PPSLOCK
	movlw 0x55			; Required sequence, next 4 instructions
	movwf PPSLOCK,BANKED
	movlw 0xAA
	movwf PPSLOCK,BANKED
	bcf PPSLOCK,0,BANKED		; Only a BCF instruction will work

	BANKSEL RX1PPS		; set pins for USART RC6 RC7
	movlw b'00010111'
	movwf RX1PPS,BANKED
	movlw b'00001001'
	movwf RC6PPS,BANKED

; Bank to PPSLOCK register
	BANKSEL PPSLOCK
	movlw 0x55			; Required sequence, next 4 instructions
	movwf PPSLOCK,BANKED
	movlw 0xAA
	movwf PPSLOCK,BANKED
	bsf PPSLOCK,0,BANKED		; Only a BSF instruction will work

Start	call setlStart
	movf RC1REG,W		; flush RS232 buffer
	movf RC1REG,W

	if TestMode == 0
	  clrf PORTD			; clear switch input reg
	  movlw 0xFF			; all inputs
	  movwf TRISD
	  movlw 0xFF
	  movwf PORTA			; LED columns HI
	  clrf PORTE			; ready to drive switch data
	  movlw b'11111110'
	  movwf TRISE			; RE0 = OUTPUT LO
	  call setlStart

	  movf PORTD,W			; read CTRL switch state
	  movwf TempA
	  nop
	  btfss TempA,ctrsLA
	  goto waitBoot1
	endif

	goto StartCode
;
; ---------------------------------------
; SUBROUTINE: DELAY - startup settle time
; ---------------------------------------
;
setlStart	clrf TempA
	clrf TempB
settle	decfsz TempA
	goto settle
	decfsz TempB
	goto settle
	return
;
; ----------------------------------
; SUBROUTINE: BLANK THE DISPLAY LEDS
; ----------------------------------
;
BlankDisp	clrf PORTD			; all LED anodes LOW
	return
;
; ----------------------------------------
; SUBROUTINE: RECEIVE CHARACTER FROM RS232 - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; ----------------------------------------
; This routine does not return until a character is received.
;
Receive	
	if TestMode == 1
	  movlw 0h
	  movwf TempA
	  return
	endif
	BANKSEL PIR3		; .receive
RXwait	btfss PIR3,RC1IF,BANKED		; (5) check for received data
	goto RXwait

	movf RC1REG,W
	movwf TempA
	return
;
; ------------------------------------------------
; SUBROUTINE: WAIT UNTIL RS232 IS FINISHED SENDING - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; ------------------------------------------------
;
TransWt	
	if TestMode == 1
	  return
	endif

	movwf TX1REG		; .transwt
TXwait	btfss TX1STA,TRMT		; (1) transmission is complete if hi
	goto TXwait
	return
;
; --------------------------------
; SUBROUTINE: READ INTERNAL EEPROM - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; --------------------------------
; address/data set up in EE regs
RDeeprom	movlw b'00000000'		; .rdeeprom Setup Data EEPROM Access, reading
	movwf NVMCON1
	bsf NVMCON1,RD		; Issue EE Read
	movf NVMDAT,W		; W = EE_DATA
	infsnz NVMADRL
	incf NVMADRH
	return 
;
; ---------------------------------
; SUBROUTINE: WRITE INTERNAL EEPROM - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; ---------------------------------
; address/data set up in EE regs
WTeeprom	movwf NVMDAT
inWTeeprom	movlw b'00000100'		; .wteeprom Setup Data EEPROM Access, enable writes
	movwf NVMCON1
	
	movlw 0x55
	movwf NVMCON2
	movlw 0xAA
	movwf NVMCON2
	bsf NVMCON1,WR		; Set WR bit to begin write

WtIEEPwt	btfsc NVMCON1,WR		; Wait for write to complete
	goto WtIEEPwt

c	infsnz NVMADRL
	incf NVMADRH
	return 
;
; --------------------------
; KEYDOWN POWERUP BOOT ENTRY 
; --------------------------
; PC will send a go Boot CMD when wanting to talk to boot loader
; Same protocol as from main loop
waitBoot1	call BlankDisp
	movlw 'B'
	call TransWt
	movlw 't'
	call TransWt
	
waitBoot2	call Receive
	xorlw rzMainCmdGoProg
	btfss STATUS,Z
	goto waitBoot2
;
; -----------------------------------------
; NEW BOOT CODE ROM AREA - FOR UPDATING ROM
; -----------------------------------------
;
doBootCode	call BlankDisp
	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	clrf TBLPTRU
	movlw HIGH ProgramCode
	movwf TBLPTRH
	movlw LOW ProgramCode
	movwf TBLPTRL
	movlw rzMainCmdGoProg		; ACK prog ROM
TXloop	call TransWt
;
; ----------------------
; WAITING FOR PC COMMAND
; ----------------------
;
WaitCom	movlw HIGH WaitCom
	movwf PCLATH
	call Receive		; wait and read from RS232
	andlw 0x07
	mullw 4h			; PCL offset for GOTO
	movf PRODL,W
	addwf PCL
	goto SetAddr		; change address		(0)
	goto WaitCom		;			(1)
	goto FillBuff		; fill buffer		(2)
	goto WriteProg		; write the buffer data		(3)
	goto BTRDeep		; read internal EEPRROM		(4)
	goto BTWTeep		; write internal EEPROM		(5)
	goto WaitCom		;			(6)
	goto EndProg		; finished with prog ROM	(7)
;
; ------------------
; CHANGE ROM ADDRESS
; ------------------
;
SetAddr	call Receive		; ROM addr U
	movwf TBLPTRU
	call Receive		; ROM addr h
	movwf TBLPTRH
	call Receive		; ROM data L
	movwf TBLPTRL

; make sure address is not < 0x000300

	movf TBLPTRU
	btfss STATUS,Z
	goto AddrOK

	movlw 0x02
	cpfsgt TBLPTRH
	goto AddrFail

AddrOK	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L

	movlw rzBootCmdAddr		; ACK
	goto TXloop

AddrFail	movlw 0xFF
	goto TXloop
;
; -------------------
; FILL PROGRAM BUFFER
; -------------------
;
FillBuff	movlw d'4'			; 4 bytes at a time - no Xtal for OSC - minimising data transfer size
	movwf TempB
FBloop	call Receive
	movwf POSTINC0
	decfsz TempB
	goto FBloop
 
	movlw rzBootCmdFill
	goto TXloop
;
; -----------------------
; PROGRAM BUFFER INTO ROM
; -----------------------
;
WriteProg	SetReadROM			; test if ROM is same as program data, no program if so
	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	movf TBLPTRU,W
	movwf TempA
	movf TBLPTRH,W
	movwf TempB
	movf TBLPTRL,W
	movwf TempC
	movlw d'128'
	movwf TempD

verfLoop	tblrd*+
	movf TABLAT,W
	xorwf POSTINC0,W
	btfss STATUS,Z
	goto notVerf

	decfsz TempD
	goto verfLoop

	call IncAddr
	movlw rzBootCmdNoProg		; ACK no prog change
	goto TXloop
;
; now do programming
;
notVerf	movf TempA,W
	movwf TBLPTRU
	movf TempB,W
	movwf TBLPTRH
	movf TempC,W
	movwf TBLPTRL

	movlw b'10010100'		; erase block
	movwf NVMCON1
	nop
	movlw 0x55
	movwf NVMCON2
	movlw 0xAA
	movwf NVMCON2
	bsf NVMCON1,WR		; start erase (CPU stalls)

	tblrd*-			; set back by 1 - ensures that upper bit still access proper write block
	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	movlw d'128'
	movwf TempD
btProgLoop	movf POSTINC0,W		; get low byte of buffer data
	movwf TABLAT		; present data to table latch
	tblwt+*			; write data, perform a short write to internal TBLWT holding register.
	decfsz TempD		; loop until holding registers are full
	goto btProgLoop

	movlw b'10000100'		; write block
	movwf NVMCON1
	nop
	movlw 0x55
	movwf NVMCON2
	movlw 0xAA
	movwf NVMCON2
	bsf NVMCON1,WR		; start write (CPU stalls)

	call IncAddr
	movlw rzBootCmdProg		; ACK
	goto TXloop
;
; ------------------------------
; INCREMENT ADDRESS BY 128 BYTES
; ------------------------------
;
IncAddr	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L

	movf TempA,W
	movwf TBLPTRU
	movf TempB,W
	movwf TBLPTRH
	movf TempC,W
	movwf TBLPTRL

	movlw 0x80
	addwf TBLPTRL
	btfss STATUS,C
	return

	infsnz TBLPTRH
	incf TBLPTRU
	return
;
; -----------------------
; FINISHED WITH BOOT CODE
; -----------------------
;
EndProg	bcf NVMCON1,WREN
	goto Start			; do a restart in main code
;
; ---------------------
; READ INTERNAL EEPROM
; ---------------------
;
BTRDeep	call Receive
	movwf NVMADRH
	call Receive
	movwf NVMADRL
	movlw rzBootCmdEEread		; ACK read
	call TransWt
	call RDeeprom		; data
	goto TXloop
;
; ---------------------
; WRITE INTERNAL EEPROM
; ---------------------
;
BTWTeep	call Receive
	movwf NVMADRH
	call Receive
	movwf NVMADRL
	call Receive
	call WTeeprom		; write EEPROM routine
	movlw rzBootCmdEEwrit		; ACK - done
	goto TXloop

; ------------------------------------------------------------------------------
;	END OF PROTECTED AREA
; ------------------------------------------------------------------------------

	org 0x0300

ProgramCode
;
; -------------------
; PC COMMS JUMP TABLE
; -------------------
;
PC_Comms	btfsc genBitsWIF,wifSerial	; no PC serial if WIF serial active
	goto CodeExec

	BANKSEL PIR3
	btfss PIR3,RC1IF,BANKED		; (5) check for PC command
	goto CodeExec		; none

	btfsc rfdeLEDs,Run
	call ClearRUN		; stop run mode

	movlw HIGH PC_Comms
	movwf PCLATH
	movf RC1REG,W		; wait and read from RS232
	andlw 0x0F
	mullw 4h			; PCL offset for GOTO
	movf PRODL,W
	addwf PCL
	goto EDUCcon		; 0
	goto doBootCode
	goto pcMemWT
	goto pcMemRD
	goto pcMemER
	goto pcOptions
	goto pcPrgNames
	goto pcWritEEP
	goto pcReadEEP		; 7
	goto pcErasEEP		; 8
	goto doBootCode2
	goto MainLoop
	goto MainLoop
	goto MainLoop
	goto MainLoop
	goto pcTest			; F
;
; ----------------------------------------------
; SUBROUTINE: set OPR LEDs from instruction code
; ----------------------------------------------
;  - in lower 4 bits [3 2 1 x]
getOPRLED	movlw HIGH getOPRLED
	movwf PCLATH
	movf TempB,W
	andlw 0x0E
	addwf PCL
	retlw oprAND
	retlw oprTAD
	retlw oprISZ
	retlw oprDCA
	retlw oprJMS
	retlw oprJMP
	retlw oprIOT
	retlw oprOPR
;
; modify input switches to match screwed up inputs
;
StartCode	clrf NVMADRH		; internal EEPROM 0x000
	clrf NVMADRL		; bits 1 0 - 00 = 00, 01 = FF, 1x = Random
	call RDeeprom
	call doUserFlags

	clrf insLEDs
	clrf genBits1
	clrf genBits2
	clrf genBitsWIF
	clrf statusWIF
	clrf errorFlags
	clrf LEDupdate
	clrf ScanDigit
	clrf binSwitch
	clrf ctrSWold
	clrf ctrSWnew
	clrf ctrSWuse
	clrf fgsExecDefer
	clrf ROMaddrH
	clrf ROMaddrL
	clrf intEEPaddr
	movlw swScanRate
	movwf SWscan
	movlw SwichScanRate		; 30mS
	movwf SwRdDelay
	movlw d'50'			; default 500mS TIMER delay
	movwf delTIMER
	call ClearRUN

	movlw b'10000000'		; 48uS approx cycle time
	movwf T0CON0
	movlw b'01000010'
	movwf T0CON1

	if TestMode == 1
	  call educFirst
	  bsf rfdeLEDs,Run
	  bsf genBits1,fgSlowFast
	  movlw funcFetch
	  movwf execFunction
	  bsf genBits1,fgSinCon
	  bcf rfdeLEDs,ResetT1T13 ; - reset on T13
	  bcf rfdeLEDs,wiWIF ; NRM or WIF mode
	  clrf PCregL
	  clrf PCregH
	endif
;
; --------------
; MAIN CODE LOOP
; --------------
; Master clock cycle is 500KHz
; 24 cycles per instruction = 2uS * 24 = 48uS
; Fetch/Execute takes 96uS (10KHz), Fetch/Defer/Execute takes 120uS (8.3KHz)
; PIC clock @ 32MHz = 384 instructions per 48uS
; Update the LEDs every 51 x 48uS loops ~2.5mS
; Scan keyboard every 30mS = 12 x LEDscan rate
; SLOW mode = programmable (x) seconds per cycle - default = 500mS ($28B0 10,416 x 48uS loops)
;
MainLoop
	BANKSEL PIR0
MLwait	btfss PIR0,TMR0IF,BANKED
	goto MLwait

	bcf PIR0,TMR0IF,BANKED
	movf MainDelay,W
	movwf TMR0L

	btfss PIR4,TMR1IF,BANKED	; PIR4 in same bank as PIR0
	goto noTIMER

	decfsz delTIMER		; X x 10mS delay
	goto rstTIMER

	bsf genBits2,fgTimer		; if TMR1 overflow flag set then set TIMER flag
	movf valTIMER,W		; value from TIMER instruction
	movwf delTIMER		; reset counter timer
rstTIMER	bcf PIR4,TMR1IF,BANKED
	movlw 0x64
	movwf TMR1H
	movlw 0x00
	movwf TMR1L

noTIMER	decfsz LEDupdate
	goto CodeExec

	call ScanKeyLED
	movf LEDupCnt,W
	movwf LEDupdate
	goto PC_Comms
;
; ----------------------------------
; SUBROUTINE:  SET USER EDUC-8 FLAGS
; ----------------------------------
; Data in NVMDAT
doUserFlags	movlw maskClrFlags	
	andwf rfdeLEDs
	btfss NVMDAT,EEflgWIF
	bsf rfdeLEDs,wiWIF		; user flag set for WIF mode
	btfss NVMDAT,EEflgZPM
	bsf rfdeLEDs,PageZero		; user flag set for Zero Page mode
	btfss NVMDAT,EEflgSED
	bsf rfdeLEDs,SlowExamDep	; user flag set for allow slow Examine/Deposit
	btfss NVMDAT,EEflgT13
	bsf rfdeLEDs,ResetT1T13		; user flag set for T1 reset
	call SetCalRate
	movlw 0xE1			; WIF faster mainloop
	btfss rfdeLEDs,wiWIF
	movlw 0xA1			; NRM slower mainloop
	movwf MainDelay
	movlw d'33'			; WIF led display update
	btfss rfdeLEDs,wiWIF
	movlw d'15'			; NRM led display update
	movwf LEDupCnt
	call TMR1SET		; enable TMR1 if WIF mode
	btfsc NVMDAT,EEflgRst
	call EraseMem		; clear all registers
	btfss NVMDAT,EEflgRst
	call DefErase		; set some registers to a default state
	return
;
; -------------------------------------------
; SUBROUTINE: SET CAL RATE FOR SLOW RUN SPEED
; -------------------------------------------
;
SetCalRate	movlw SlowMinH
	movwf SlowCalH
	movlw SlowMinL
	movwf SlowCalL
	swapf NVMDAT,W
	andlw 0x03
	addlw 1h
	movwf TempA
slowCalLP	bcf STATUS,C
	dcfsnz TempA
	return
	bcf STATUS,C
	rlcf SlowCalL
	rlcf SlowCalH
	goto slowCalLP
;
; -----------------------------------------------
; SUBROUTINE: EXECUTE FRONT PANEL SWITCH FUNCTION
; -----------------------------------------------
;
PanelFunc	call BlankDisp
	clrf NVMADRH
	clrf NVMADRL
	call RDeeprom
	btfsc binSwitch,4
	goto funcPGM		; load/save programs when SW4 = 1

; SW4 = 0 = set options
; SW6   SW5
;  0     0  Set Normal or WIF operation
;  0     1  Set / Clear Slow Exam Deposit Mode
;  1     0  Set T1 or T13 Reset
;  1     1  Set / Clear Zero Page Mode

	btfsc binSwitch,6
	goto funcOpt1x

	btfsc binSwitch,5
	goto funcOpt01

	bsf NVMDAT,EEflgWIF		; Normal
	btfss binSwitch,7		; 1 = Norm, 0 = WIF
	bcf NVMDAT,EEflgWIF		; WIF
	goto funcOPTset

funcOpt01	bsf NVMDAT,EEflgSED		; Normal
	btfss binSwitch,7		; 1 = Norm, 0 = SED
	bcf NVMDAT,EEflgSED		; SED
	goto funcOPTset

funcOpt1x	btfsc binSwitch,5
	goto funcOpt11

	bsf NVMDAT,EEflgT13		; Normal
	btfss binSwitch,7		; 1 = Norm, 0 = T1
	bcf NVMDAT,EEflgT13		; T1
	goto funcOPTset

funcOpt11	bsf NVMDAT,EEflgZPM		; Normal
	btfss binSwitch,7		; 1 = Norm, 0 = ZPM
	bcf NVMDAT,EEflgZPM		; ZPM

funcOPTset	clrf NVMADRH
	clrf NVMADRL
	movf NVMDAT,W
	call WTeeprom
	call doUserFlags
	bsf errorFlags,efgERRdisplay	; flash PCregL LED[0] once
	bsf errorFlags,efgOK
	call ErrorStart
	return
;
; FRONT PANEL - LOAD/SAVE PROGRAM
;
funcPGM	btfsc binSwitch,7
	goto LoadProg

	btfss binSwitch,6
	goto SaveWIF

; -------------------------------------------
SaveNRM	movf binSwitch,W		; set external eeprom address
	andlw 0x07
	movwf EEaddrH
	movwf TempA			; save for prog num text
	clrf EEaddrL
	call EDnumToEEP		; get internal eeprom name address

	movlw 0x08			; write 8 x 32 bytes = 256 bytes
	goto WritEEprg

; -------------------------------------------
SaveWIF	movf binSwitch,W		; if valid set external eeprom address
	andlw 0x03
	movwf EEaddrH
	xorlw 0x03
	btfsc STATUS,Z
	clrf EEaddrH		; if 3 set on switches -> address = 0
	nop
	movf EEaddrH,W
	movwf TempA			; save for prog num text
	call WInumToEEP		; get internal eeprom name address

	movlw 0x08			; WIF EE address = EDUC8 prog space + (WIF address * 8)
	mulwf EEaddrH
	movlw 8h			; add offset for EDUC-8 program storage
	addwf EEaddrH
	clrf EEaddrL

	movlw d'64'			; write 64 x 32 bytes = 2048 bytes
WritEEprg	movwf TempD

	movlw MEMbuff		; load data into memory
	movwf FSR0H
	clrf FSR0L
writePNm	call wrtEEP32
	btfsc genBits2,EEerror
	goto pgmEEPerr

	decfsz TempD
	goto writePNm

	clrf TempB			; write options byte
	; bcf TempB,pibValid		; valid - already clear
	btfss rfdeLEDs,PageZero
	bsf TempB,pibZPM		; Normal
	btfss rfdeLEDs,SlowExamDep
	bsf TempB,pibSED		; Normal
	btfss rfdeLEDs,ResetT1T13	; ResetT1T13 = 1, Reset on T1 or BO
	bsf TempB,pibT1T13		; T13
	nop
	movf TempB,W
	call WTeeprom

	clrf TempB			; write "program " + program number
pgNameLoop	call BlankName
	xorlw 0xFF
	btfsc STATUS,Z
	goto pgNameEnd

	xorlw 0xFF
	call WTeeprom
	incf TempB
	goto pgNameLoop

pgNameEnd	call ProgNumTxt		; data in TempA
	call WTeeprom

	movlw d'14'			; Program_X 9 chars - need 14 blanks
	movwf TempB			; write "program " + program number
pgBkName	movlw " "
	call WTeeprom
	decfsz TempB
	goto pgBkName

	goto pgmRWok		; done

; -------------------------------------------
LoadProg	btfss binSwitch,6
	goto LoadWIF

; -------------------------------------------
LoadNorm	bcf genBits2,fgTemp		; Normal mode
	movf binSwitch,W		; get address (0-7) (binSwitch and 7)
	andlw 0x07
	call EDnumToEEP		; get internal eeprom name address
	call RDeeprom		; load info byte
	btfsc NVMDAT,pibValid		; if not valid (=1) then flash PCregL LED[0]
	goto pgmInvalid

	bcf genBits2,fgTemp		; Normal mode
	movf binSwitch,W		; if valid set external eeprom address
	andlw 0x07
	movwf EEaddrH
	clrf EEaddrL

	clrf TempC
	movlw 0x01			; 1 256 page (+1) 256 bytes
	goto ReadEEprg

; -------------------------------------------
LoadWIF	bsf genBits2,fgTemp		; WIF mode
	movf binSwitch,W		; if valid set external eeprom address
	andlw 0x03
	movwf EEaddrH
	xorlw 0x03
	btfsc STATUS,Z
	clrf EEaddrH		; if 3 set on switches -> address = 0
	nop
	movf EEaddrH,W
	call WInumToEEP		; get internal eeprom name address
	call RDeeprom		; load info byte
	btfss NVMDAT,pibValid		; if not valid (=1) then flash PCregL LED[0]
	goto WIFvalid

pgmInvalid	bsf errorFlags,efgERRdisplay	; flash PCregL LED[0] continous
	bsf errorFlags,efgError01
	call ErrorStart
	return
	
WIFvalid	bsf genBits2,fgTemp		; WIF mode
	movlw 0x08			; WIF EE address = EDUC8 prog space + (WIF address * 8)
	mulwf EEaddrH
	movlw 8h			; add offset for EDUC-8 program storage
	addwf EEaddrH
	clrf EEaddrL

	clrf TempC
	movlw 0x09			; 8 256 pages (+1) 2048 bytes
ReadEEprg	movwf TempD

	movlw MEMbuff		; load data into memory
	movwf FSR0H
	clrf FSR0L

	call StartRD
	btfsc genBits2,EEerror
	goto pgmEEPerr
pgmRDloop	call EEread
	movwf POSTINC0
	decfsz TempC
	goto pgmRDloop
	decfsz TempD
	goto pgmRDloop

	bcf genBits2,n_ack		; done
	call EEread

	movlw maskClrFlags		; set operating config from info byte	
	andwf rfdeLEDs
	btfsc genBits2,fgTemp
	bsf rfdeLEDs,wiWIF		; WIF mode
	btfss NVMDAT,pibZPM
	bsf rfdeLEDs,PageZero
	btfss NVMDAT,pibSED
	bsf rfdeLEDs,SlowExamDep
	btfss NVMDAT,pibT1T13
	bsf rfdeLEDs,ResetT1T13		; reset T1

pgmRWok	bsf errorFlags,efgOK
	bsf errorFlags,efgERRdisplay
	call ErrorStart
	return
;
; EXTERNAL EEPROM ERROR OCCURRED FROM INTERNAL PIC ACCESS
;
pgmEEPerr	bsf errorFlags,efgERRdisplay	; flash PCregL LED[7][0] continous
	bsf errorFlags,efgError71
	call ErrorStart
	return

	org 0x0700
;
; --------------------------------------
; SUBROUTINE: GET INSTRUCTION SRCTGT LED
; --------------------------------------
;
GetST_LED	movwf TempD
	movlw HIGH GetST_LED
	movwf PCLATH
	bcf STATUS,C
	rlcf TempD,W
	addwf PCL
	retlw wstAC
	retlw wstREG
	retlw wstI
	retlw wstPC
;
; ---------------------
; SUBROUTINE: ~5uS DELAY
; ---------------------
;
CKdelay	movlw 8h
	movwf DelA
CKdelayLP	nop
	decfsz DelA
	goto CKdelayLP
	return
;
; ----------------------------------------------
; SUBROUTINE: SET BIT MASK FOR WIF BSF BCF TYPES
; ----------------------------------------------
; bbbr rrrr
wiBitMask	swapf wiCodeL,W
	andlw 0x0F
	movwf TempC
	bcf STATUS,C
	rrcf TempC
	call getBitMask
	return
;
; ---------------------------
; WIF MODE INSTRUCTION DECODE
; ---------------------------
;
WIFins	btfsc genBitsWIF,wifNewINS
	goto DecodeWIF

	movf PCregH,W	; noTest
	andlw wiRomHI
	movwf FSR0H
	movf PCregL,W
	movwf FSR0L
	bcf STATUS,C		; WIF memory has 2 bytes
	rlcf FSR0L
	rlcf FSR0H
	movlw MEMbuff
	addwf FSR0H
	movf POSTINC0,W
	movwf wiCodeH		; ROM word H
	movf POSTINC0,W
	movwf wiCodeL		; ROM word L
	call SetInsLEDs

	bsf genBitsWIF,wifNewINS
	movf wiCodeH,W		; instruction index
	andlw b'00001100'		; yyyyxx00 00000000
	movwf wifADDR		; 00000000 00xxyyyy
	swapf wiCodeH,W
	rlncf wifADDR
	rlncf wifADDR
	andlw 0x0F
	iorwf wifADDR,W
	movwf wifADDR
	mullw 4h			; create table offset
	movf PRODL,W
	movwf wifTblMain

	movlw picADDR		; table offset 01
	subwf wifADDR,W
	mullw 4h
	movf PRODL,W
	movwf wifTbl013

	bcf genBitsWIF,wifSkip
	movlw picGOSUB		; table offset 10
	subwf wifADDR,W
	mullw 4h
	movf PRODL,W
	movwf wifTbl1034

	bcf genBitsWIF,wifDev		; dev 1
	movf wiCodeL,W
	andlw WIdevSelect		; 0 = dev 0, 1 = dev 1
	btfss STATUS,Z
	bsf genBitsWIF,wifDev		; dev 1
	goto DecodeWIF
;
; ----------------------------------
;SUBROUTINE: WIF SELECT INPUT DEVICE
; ----------------------------------
;
wifSetIOin	btfsc genBitsWIF,wifDev
	goto widevIN1
	
	selIOin0			; select CLK/RstOUT DEV 0 IN
	return
widevIN1	
	selIOin1			; select CLK/RstOUT DEV 1 IN
	return
;
; -----------------------------------
;SUBROUTINE: WIF SELECT OUTPUT DEVICE
; -----------------------------------
;
wifSetIOout	btfsc genBitsWIF,wifDev
	goto widevOUT1
	
	selIOout0			; select CLK/RstOUT DEV 0 OUT
	return
widevOUT1	
	selIOout1			; select CLK/RstOUT DEV 1 OUT
	return
;
; ----------------------------
; SUBROUTINE: SET SRC/TGT LEDS
; ----------------------------
;
SetInsLEDs	movf wiCodeH,W		; ROM word H  0000 00ss
	andlw 0x03
	call GetST_LED		; SRC TGT LEDs
	movwf insLEDs
	rlncf wiCodeL,W		; ROM word L  tt00 0000
	movwf TempD
	rlncf TempD,W
	andlw 0x03
	call GetST_LED		; SRC TGT LEDs
	iorwf insLEDs		; light LEDs for source, dest
	return

	org 0x0800
;
; ---------------------------
; JUMP TO MEM REF INS CYCLE 1
; ---------------------------
;
MRdecode1	movlw HIGH MRdecode1
	movwf PCLATH
	bcf STATUS,C
	rlcf TempB,W
	andlw 0x1E
	addwf PCL
	goto mriAND1
	goto mriTAD1
	goto mriISZ1
	goto mriDCA1
	goto mriJMS1
	goto mriJMP1
	goto mriIOT1
	goto mriAUG1
;
; -------------------------------
; JUMP TO MEM REF INS CYCLE 2 - 9
; -------------------------------
;
MRdecode29	movlw HIGH MRdecode29
	movwf PCLATH
	bcf STATUS,C
	rlcf TempB,W
	andlw 0x1E
	addwf PCL
	goto mriAND29
	goto mriTAD29
	goto mriISZ29
	goto mriDCA29
	goto mriJMS29
	goto mriJMP29
	goto mriIOT29
	goto mriAUG29
;
; ----------------------------
; JUMP TO MEM REF INS CYCLE 13
; ----------------------------
;
MRdecode13	movlw HIGH MRdecode13
	movwf PCLATH
	bcf STATUS,C
	rlcf TempB,W
	andlw 0x1E
	addwf PCL
	goto mriAND13
	goto mriTAD13
	goto mriISZ13
	goto mriDCA13
	goto mriJMS13
	goto mriJMP13
	goto mriIOT13
	goto mriAUG13
;
; -------------------------------
; JUMP TO MEM REF INS CYCLE 14-21
; -------------------------------
;
MRdecod1421	movlw HIGH MRdecod1421
	movwf PCLATH
	bcf STATUS,C
	rlcf TempB,W
	andlw 0x1E
	addwf PCL
	goto mriAND1421
	goto mriTAD1421
	goto mriISZ1421
	goto mriDCA1421
	goto mriJMS1421
	goto mriJMP1421
	goto mriIOT1421
	goto mriAUG1421
;
; ------------------------
; SUBROUTINE: LED/KEY SCAN
; ------------------------
;
ScanKeyLED	movlw 0xFF			; displays/key COLs off
	movwf PORTA			; SCL and SDA in PORTA but already in STOP condition

	movlw HIGH LedJT
	movwf PCLATH
	movf ScanDigit,W
LedJT	addwf PCL
	goto ROW_LK0
	goto ROW_LK1
	goto ROW_LK2
	goto ROW_LK3
	goto ROW_LK4
	goto ROW_LK5
	goto ROW_LK6

ROW_LK0	movf PCregL,W		; PC LEDs set
	movwf PORTD

	movlw b'11111110'		; first column
	movwf PORTA
	movlw 4h
	addwf ScanDigit
	return

ROW_LK1	movf MAreg,W		; MA LEDs set
	movwf PORTD

	movlw b'11111101'		; next column
	movwf PORTA
	movlw 4h
	addwf ScanDigit
	return

ROW_LK2	movf MBreg,W		; MB LEDs set
	movwf PORTD

	movlw b'11111011'		; next column
	movwf PORTA
	movlw 4h
	addwf ScanDigit
	return

ROW_LK3	movf ACreg,W		; AC LEDs set
	movwf PORTD

	movlw b'11110111'		; next column
	movwf PORTA
	movlw 4h
	addwf ScanDigit
	return

ROW_LK4	movf insLEDs,W		; Instruction LEDs set
	movwf PORTD

	movlw b'11101111'		; next column
	movwf PORTA
	movlw 4h
	addwf ScanDigit
	return

ROW_LK5	movf rfdeLEDs,W		; run/fetch/etc LEDs set
	iorwf fgsExecDefer,W
	movwf PORTD
	clrf fgsExecDefer

noFGSed	movlw b'11011111'		; next column
	movwf PORTA

	decfsz SwRdDelay
	goto NoSWyet

	movlw SwichScanRate		; ~10mS, switch debounce
	movwf SwRdDelay

	movlw 4h
	addwf ScanDigit
	return

NoSWyet	clrf ScanDigit
	return

ROW_LK6	
	if TestMode == 1
	  goto exitLK6
	endif

; -------------------------------------
	clrf PORTD			; clear switch input reg
	movlw 0xFF			; all inputs
	movwf TRISD
	movlw 0xFF
	movwf PORTA			; LED columns HI
	movlw b'11111110'
	movwf TRISE			; RE0 = OUTPUT LO
	call SMdelay

; -------------------------------------
	movf PORTD,W		; read CTRL switch state
	andlw b'01111111'
	movwf ctrSWnew

	movf PORTB,W		; merge external          LA(7) DEP(6)
	andlw b'11000000'
	movwf TempD			; with front panel switch LA(6) DEP(5) states
	nop
	rrncf TempD,W		; internal LA and DEP switches are normally in logic 1 state, = LO when switched active
	iorlw b'00011111'
	andwf ctrSWnew

; -------------------------------------
exWIF1	call SMdelay
	movlw b'11111101'
	movwf TRISE			; RE1 = OUTPUT LO
	call SMdelay

	movf PORTD,W		; read BINARY switch state
	movwf TempD
	movlw 8h
	movwf TempC
adjSW	rrcf TempD
	rlcf binSwitch
	decfsz TempC
	goto adjSW

; -------------------------------------
	movlw b'11111011'		; testcode - reenable for new PCB
	movwf TRISE			; RE2 = OUTPUT LO, RE1 RE0 inputs, enable 244 chip outputs
	call CKdelay

	movf PORTD,W		; read DEV 0 Parallel Input switch connects
	andwf binSwitch		; external switches are active low - EDUC-8 binary switches should be in Logic 1 position to work

	movlw b'11111111'		; RE012 = INPUTs, 244 chip outputs HiZ
	movwf TRISE
	nop
	clrf TRISD			; PORTD = OUTPUTs
;
; if first switches read, set initial state of switch registers
;
	btfsc genBits2,InitSW		; if = 0, 1st switch reads - set initial switch registers
	goto exitLK6

	bsf genBits2,InitSW
	movf ctrSWnew,W
	movwf ctrSWold
	call swSinCon		; sets initial state of switch flags
	call swSlwFst

exitLK6	clrf ScanDigit
	call SWfunction
	return
;
; ---------------------------------
; SUBROUTINE: RETURN BIT 0 - 7 MASK
; ---------------------------------
; val 0 - 7 in TempC
getBitMask	movlw HIGH getBitMask
	movwf PCLATH
	rlncf TempC,W
	addwf PCL
	retlw b'00000001'
	retlw b'00000010'
	retlw b'00000100'
	retlw b'00001000'
	retlw b'00010000'
	retlw b'00100000'
	retlw b'01000000'
	retlw b'10000000'
;
; -------------------------------------------------
; SUBROUTINE: READ EEPROM WITH NO ADDRESS INCREMENT
; -------------------------------------------------
;
RDeepromNI	movlw b'00000000'		; .rdeeprom Setup Data EEPROM Access, reading
	movwf NVMCON1
	bsf NVMCON1,RD		; Issue EE Read
	movf NVMDAT,W		; W = EE_DATA
	return
;
; ----------------------------------------------------------------
; SUBROUTINE: INSTRUCTION SEROUT/SERIN RESET SERIAL PORT IF ERRORS
; ----------------------------------------------------------------
;
Rst232	movf RC1STA,W
	andlw b'00000110'		; FERR OERR flags
	btfsc STATUS,Z
	return

	call BlankDisp
	bcf RC1STA,SPEN		; disable Comms - clear FERRs etc
	call do60WT
	bsf RC1STA,SPEN		; renable Comms
	movf RC1REG,W
	movf RC1REG,W
	return

	org 0x0A00
;
; PC SET MEMORY TO A VALUE
;
pcMemWT	call BlankDisp
	movlw MEMbuff
	movwf FSR0H
	clrf FSR0L

writMMREQ	movlw rzMainCmdMemWT
	call TransWt

	call Receive		; new value available?
	xorlw pcEEPend
	btfss STATUS,Z
	goto nwMemData		; yes

endFMV	movlw pcEEPend		; no, finished
	call TransWt
	goto MainLoop

nwMemData	movlw 4h
	movwf TempB
nwMMlp	call Receive
	movwf POSTINC0
	decfsz TempB
	goto nwMMlp
	goto writMMREQ
;
; PC ERASE MEMORY
;
pcMemER	call EraseMem
	movlw pcEEPend
	call TransWt
	goto MainLoop
;
; PC READ MEMORY
;
pcMemRD	call BlankDisp
	movlw MEMbuff
	movwf FSR0H
	clrf FSR0L

inRDmemRQ	movlw rzMainCmdMemRD
	call TransWt

readMMREQ	call Receive		; does PC want data
	xorlw pcEEPend
	btfss STATUS,Z
	goto doMemRead		; yes

	movlw pcEEPend		; no, ACK finished
	call TransWt
	goto MainLoop

doMemRead	movf TempA,W
	xorlw pcEEPread
	btfss STATUS,Z
	goto MainLoop		; error - exit - PC will timeout

	movlw pcEEPread
	call TransWt
	movlw d'4'			; send 4 bytes
	movwf TempB

mmXferLoop	movf POSTINC0,W
	call TransWt
	decfsz TempB
	goto mmXferLoop
	goto readMMREQ
;
; PC Read/Write options
;
pcOptions	call BlankDisp
	clrf NVMADRH
	clrf NVMADRL
	movlw rzMainCmdOpts
	call TransWt

	call Receive
	xorlw pcEEPwrite
	btfss STATUS,Z
	goto intEEPrd

	call Receive		; data to write
	call WTeeprom
	call doUserFlags
	movlw pcEEPwrite
	call TransWt
	goto MainLoop

intEEPrd	movf TempA,W
	xorlw pcEEPread
	btfss STATUS,Z
	goto MainLoop		; CMD error

	movlw pcEEPread
	call TransWt
	call RDeeprom		; get slow speed mode
	andlw b'00110000'
	movwf TempB
	btfss rfdeLEDs,wiWIF
	bsf TempB,EEflgWIF		; Normal
	btfss rfdeLEDs,PageZero
	bsf TempB,EEflgZPM		; Normal
	btfss rfdeLEDs,SlowExamDep
	bsf TempB,EEflgSED		; Normal
	btfss rfdeLEDs,ResetT1T13
	bsf TempB,EEflgT13		; T13
	nop
	movf TempB,W
	call TransWt
	goto MainLoop
;
; ------------------------------
; PC ASKED IF EDUC8 IS CONNECTED
; ------------------------------
;
EDUCcon	movlw 0xED			; send EDUC-8 online to PC
	call TransWt
	movlw 0xC8
	call TransWt
	movf NVMDAT,W		; will have eeprom user setting data
	call TransWt

	; send other data if required - *** remember to update help file PC protocol ***
	goto MainLoop
;
; ---------------------------------------------------
; SUBROUTINE: IF FERR OR OERR THEN RESET SERIAL COMMS
; ---------------------------------------------------
; When USB or PC progFM disconnects - PIC may see continuous 0's
; reste RS232 to clear the error
;
RstRS232	movf RC1STA,W
	andlw b'00000110'		; FERR OERR flags
	btfsc STATUS,Z
	goto flushRX

	bcf RC1STA,SPEN		; disable Comms - clear FERRs etc
	call BlankDisp
	call do60WT
	bsf RC1STA,SPEN		; renable Comms
flushRX	movf RC1REG,W
	movf RC1REG,W
	return
;
; -----------------------
; SUBROUTINE: SMALL DELAY
; -----------------------
;
SMdelay	movlw 4h
	movwf DelA
SMwait	decfsz DelA
	goto SMwait
	return
;
; --------------------------
; SUBROUTINE: ~ 60mS @ 32MHz
; --------------------------
;
do60WT	movlw 9h
	movwf DelC
do60wtLP	call DelayWt
	decfsz DelC
	goto do60wtLP
	return
;
; -------------------------------
; SUBROUTINE: ~ 7mS DELAY @ 32MHZ
; -------------------------------
;
DelayWt	clrf DelA
	movlw 50h
	movwf DelB
DLWlp	decfsz DelA
	goto DLWlp
	decfsz DelB
	goto DLWlp
	return
;
; --------------------------
; SUBROUTINE: CLEAR RUN MODE
; --------------------------
;
ClearRUN	movlw initCycleIndex
	movwf execIndex
	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetFetch
	movwf rfdeLEDs
	clrf execFunction
	movlw 0x01			; make sure FAST/SLOW modes cycles execute straight away when started
	movwf cycleCountH
	movwf cycleCountL
	return
;
; -----------------------------------------
; DETERMINE WHAT SWITCH FUNCTION TO EXECUTE
; -----------------------------------------
;
SWfunction	
	if TestMode == 1
	  return
	endif

	btfsc genBits1,fgSwPanFunc	; Panel function executed?
	goto isswPF			; yes

	btfss ctrSWuse,ctrsHL		; Halt down
	goto normSW			; no

	btfsc ctrSWnew,ctrsLA		; LA down
	goto normSW			; no

isPFstrt	bsf genBits1,fgSwPanFunc
	call PanelFunc
	return

isswPF	btfsc ctrSWnew,ctrsLA		; LA up?
	goto swHalt			; yes, do HALT SW
	return			; no

; -----------------------------------------------
normSW	movf ctrSWnew,W
	xorwf ctrSWold,W
	movwf TempD			; non zero = switch change
	btfsc STATUS,Z
	return

	btfsc TempD,ctrsLA
	goto swLoadAddr

	btfsc errorFlags,efgERRdisplay	; only LA switch active if program load/save error condition
	goto swReturn

NoErrLA	btfsc TempD,ctrsDP
	goto swDeposit
	btfsc TempD,ctrsEX
	goto swExamine
	btfsc TempD,ctrsRN
	goto swRun
	btfsc TempD,ctrsHL
	goto swHalt
	btfsc TempD,ctrsSC
	goto swSinCon
	btfsc TempD,ctrsSF
	goto swSlwFst

swReturn	movf ctrSWnew,W
	movwf ctrSWold
	return

; ----------------------------------------

swLoadAddr	btfsc ctrSWnew,ctrsLA		; switch UP or DN?
	goto LAup

	movf ctrSWuse		; are any switches active?
	btfss STATUS,Z
	goto swReturn		; yes

	bsf ctrSWuse,ctrsLA		; switch in use
	btfss errorFlags,efgERRdisplay
	goto noLAerror

	bcf errorFlags,efgERRdisplay
	goto swReturn		; error condition

noLAerror	btfsc rfdeLEDs,Run		; execute function if not in RUN mode
	goto swReturn

	bcf genBitsWIF,wifSerial	; clear serial port active
	bcf genBits1,fgHALT		; in case set
	clrf PCregL			; LA switch down - clear PC
	clrf PCregH
	btfss rfdeLEDs,wiWIF
	goto swReturn

	clrf MAreg			; WIF mode clears these regs
	clrf MBreg
	clrf ACreg
	goto swReturn

LAup	btfss ctrSWuse,ctrsLA		; is this switch active
	goto swReturn		; no

	bcf ctrSWuse,ctrsLA		; switch not in use
	btfsc rfdeLEDs,Run		; execute LOAD ADDRESS if not in RUN mode
	goto swReturn

	movf binSwitch,W		; LA switch up = set PC from switches
	movwf PCregL
	clrf PCregH			; for WIF mode, only lower 8 bits of PC can be set from LA switch
	goto swReturn

; ----------------------------------------
swDeposit	btfsc ctrSWnew,ctrsDP		; switch UP or DN?
	goto DPup

	movf ctrSWuse		; are any switches active?
	btfss STATUS,Z
	goto swReturn		; yes

	bsf ctrSWuse,ctrsDP		; switch in use
	btfsc rfdeLEDs,Run		; execute function if not in RUN mode
	goto swReturn

	btfsc rfdeLEDs,wiWIF		; no deposit for WIF mode
	goto swReturn

	bsf rfdeLEDs,Run		; fast/slow mode - start Run
	bcf genBits1,fgHALT		; in case set
	movlw funcDeposit		; EDUC-8 enters RUN mode for DEPOSIT in SLOW mode  - Page 81
	btfsc genBits1,fgSlowFast	; if slow mode - test override flag
	goto setDEP			; fast mode

	btfss rfdeLEDs,SlowExamDep	; override set?
	movlw funcFetch		; no, start RUN mode
setDEP	movwf execFunction
	goto swReturn

DPup	bcf ctrSWuse,ctrsDP		; not in use now
	goto swReturn

; ----------------------------------------
swExamine	btfsc ctrSWnew,ctrsEX		; switch UP or DN?
	goto EXup

	movf ctrSWuse		; are any switches active?
	btfss STATUS,Z
	goto swReturn		; yes

	bsf ctrSWuse,ctrsEX		; switch in use
	btfsc rfdeLEDs,Run		; execute function if not in RUN mode
	goto swReturn

	btfsc rfdeLEDs,wiWIF		; no examine for WIF mode
	goto swReturn

	bsf rfdeLEDs,Run		; fast/slow mode - start Run
	bcf genBits1,fgHALT		; in case set
	movlw funcExamine		; EDUC-8 enters RUN mode for EXAMINE in SLOW mode  - Page 81
	btfsc genBits1,fgSlowFast	; if slow mode - test override flag
	goto setEXM			; fast mode

	btfss rfdeLEDs,SlowExamDep	; override set?
	movlw funcFetch		; no, start RUN mode
setEXM	movwf execFunction
	goto swReturn

EXup	bcf ctrSWuse,ctrsEX		; not in use now
	goto swReturn

; ----------------------------------------
swRun	btfsc ctrSWnew,ctrsRN		; switch UP or DN?
	goto RNup

	movf ctrSWuse		; are any switches active?
	btfss STATUS,Z
	goto swReturn		; yes

	bsf ctrSWuse,ctrsRN		; switch in use
	goto swReturn

RNup	bsf rfdeLEDs,Run		; RN switch up - start run
	bcf genBits1,fgHALT		; in case set
	movlw funcFetch		; execute fetch next
	movwf execFunction
RNexit	bcf ctrSWuse,ctrsRN		; not in use now
	goto swReturn

; ----------------------------------------
swHalt	btfsc ctrSWnew,ctrsHL		; switch UP or DN?
	goto HLup

	movf ctrSWuse		; are any switches active?
	btfss STATUS,Z
	goto swReturn		; yes

	bsf ctrSWuse,ctrsHL		; switch in use
	bsf genBits1,fgHALT		; will halt code if executing
	goto swReturn

HLup	btfss genBits1,fgSwPanFunc
	goto ExHaltExe

	btfsc ctrSWnew,ctrsLA		; LA switch UP?
	bcf genBits1,fgSwPanFunc	; yes, cancel sw panel function
	
ExHaltExe	bcf ctrSWuse,ctrsHL		; not in use now
	bcf genBits1,fgHALT		; as per normal EDUC-8 ops
	goto swReturn

; ----------------------------------------
swSinCon	btfss ctrSWnew,ctrsSC		; switch UP or DN?
	goto SCdn

	bcf genBits1,fgSinCon		; single
	goto swReturn

SCdn	bsf genBits1,fgSinCon		; continuous
	goto swReturn

; ----------------------------------------
swSlwFst	btfss ctrSWnew,ctrsSF		; switch UP or DN?
	goto SFdn

	bcf genBits1,fgSlowFast		; slow
	goto swReturn

SFdn	bsf genBits1,fgSlowFast		; fast
	goto swReturn

;
; ----------------------------------------
; SUBROUTINE: WRITE 32 BYTE PAGE TO EEPROM
; ----------------------------------------
;
wrtEEP32	call EEstart		; start condition
	movlw EEPwrite		; control byte write EEPROM
	call EEwrite
	btfsc genBits2,EEerror
	return
	movf EEaddrH,W
	call EEwrite		; addr HI
	btfsc genBits2,EEerror
	return
	movf EEaddrL,W
	call EEwrite		; addr LO
	btfsc genBits2,EEerror
	return

	movlw d'32'			; 24LC64 requires 32 bytes for a block write
	movwf TempC

wrtLP	movlw 0xFF
	btfss genBits1,fgErase		; erase or write  EEP bytes
	movf POSTINC0,W
	call EEwrite
	btfsc genBits2,EEerror
	return

	incf EEaddrL		; next address
	btfsc STATUS,Z
	incf EEaddrH
	decfsz TempC
	goto wrtLP

doWRTblk	call EEstop			; write the data to EEPROM
	return
;
; ---------------------------------
; SUBROUTINE: INITIATE EXT EEP READ
; ---------------------------------
;
StartRD	call EEstart		; start condition
	movlw EEPwrite		; control byte [write] EEPROM
	call EEwrite 
	btfsc genBits2,EEerror
	return
	movf EEaddrH,W		; addr HI
	call EEwrite
	btfsc genBits2,EEerror
	return
	movf EEaddrL,W		; addr LO
	call EEwrite
	btfsc genBits2,EEerror
	return
	call EEstart		; do another start condition
	movlw EEPread		; control byte [read] EEPROM
	call EEwrite
	btfsc genBits2,EEerror
	return
	bsf genBits2,n_ack		; reads have ACK
	return
;
; -------------------
; 12C START CONDITION
; -------------------
;
EEstart	bcf genBits2,EEerror
	call EEclock
	SCL_Lo			; make sure SCL = Lo
	call EEclock
	SDA_Hi			; make sure SDA = hi while SCL = Lo
	call EEclock
	SCL_Hi			; SCL = Hi
	call EEclock
	SDA_Lo			; SDA = Lo = start condition
	call EEclock
	SCL_Lo			; SCL = Lo for first clock pulse
	return
;
; ------------------
; 12C STOP CONDITION
; ------------------
;
EEstop	call EEclock
	SCL_Lo			; make sure SCL = Lo
	call EEclock
	SDA_Lo			; make sure SDA Lo while SCL = Lo
	call EEclock
	SCL_Hi			; SCL = Hi
	call EEclock
	SDA_Hi			; SDA Hi = stop condition
	call DelayWt
	call DelayWt
	return
;
; ----------------
; 12C BYTE - WRITE
; ----------------
; writes a byte of data to the EEPROM
;
EEwrite	movwf EEdat
	movlw 8h
	movwf EEcnt
EEsenLP	call EEclock
	rlcf EEdat
	btfsc STATUS,C
	goto senHI

	SDA_Lo			; Lo bit
	goto senTM

senHI	SDA_Hi			; Hi bit
senTM	call EEclock
	SCL_Hi			; clock = Hi
	call EEclock
	SCL_Lo			; clock = Lo
	decfsz EEcnt
	goto EEsenLP
;
; GET ACK FROM SLAVE
;
	SDA_In			; Data pin = input
	call EEclock		; low clock time
	SCL_Hi			; clock = Hi
	call EEclock
	SkipIfSDA_Hi		; 1 = no ack, 0 = ack
	goto ACK_OK
;
; ACK FAILED
;
	call EEstop			; finish eeprom access
	bsf genBits2,EEerror

ACK_OK	call EEclock
	SCL_Lo			; clock = Lo
	return
;
; ---------------
; 12C BYTE - READ
; ---------------
; EEread_N = no ACK after read - finished
; EEread_A = do ACK after read - continue reading
; Byte is returned in W
;
EEread	SDA_In			; SDA = input
	movlw 8h
	movwf EEcnt
EEredLP	call EEclock		; low clock count
	SCL_Hi			; clock = Hi
	call EEclock

	bcf STATUS,C
	SkipIfSDA_Low		; test SDA
	bsf STATUS,C
	rlcf EEdat
	call EEclock
	SCL_Lo			; clock = Lo
	decfsz EEcnt
	goto EEredLP		; get 8 bits
;
; READ ACK
;
	btfss genBits2,n_ack
	goto NoRDack

	SDA_Out			; SDA = output (LO)
NoRDack	call EEclock		; low clock count
	SCL_Hi			; clock = Hi
	call EEclock

	SCL_Lo			; clock = Lo
	SDA_In			; make sure SDA = In
	btfss genBits2,n_ack
	call EEstop			; finish eeprom access
	movf EEdat,W
	return
;
; -------------------------------------
; SUBROUTINE: EEPROM CLOCK SIGNAL DELAY
; -------------------------------------
;
EEclock	movlw 59h
ECloop	movwf DelA
	decfsz DelA,W
	goto ECloop
	return
;
; ------------------------------------------------------------------
; SUBROUTINE: COMPUTE EDUC INTERNAL STORAGE ADDRESS FOR PROGRAM NAME
; ------------------------------------------------------------------
;
EDnumToEEP	mullw progNameSize		; set up internal EEPROM address
	movlw LOW EDUCnames
	addwf PRODL
	btfsc STATUS,C
	incf PRODH
	movlw HIGH EDUCnames
	addwf PRODH,W
	movwf NVMADRH
	movf PRODL,W
	movwf NVMADRL
	return
;
; -----------------------------------------------------------------
; SUBROUTINE: COMPUTE WIF INTERNAL STORAGE ADDRESS FOR PROGRAM NAME
; -----------------------------------------------------------------
;
WInumToEEP	mullw progNameSize		; set up internal EEPROM address
	movlw LOW WIFnames
	addwf PRODL
	btfsc STATUS,C
	incf PRODH
	movlw HIGH WIFnames
	addwf PRODH,W
	movwf NVMADRH
	movf PRODL,W
	movwf NVMADRL
	return
;
; PC - read/write program name(s)
;
pcPrgNames	call BlankDisp		; clear display
	movlw rzMainCmdName
	call TransWt	

pgnDTread	call Receive
	xorlw pcEEPinit
	btfss STATUS,Z
	goto NMread

	call Receive		; program number
	movwf TempB
	call Receive		; program type - 0 = EDUC, FF = WIF
	xorlw 0x00
	btfss STATUS,Z
	goto izWIFpgn

; read/write EDUC 8 program name
	movf TempB,W
	call EDnumToEEP
	movlw pcEEPinit
	call TransWt
	goto pgnDTread

izWIFpgn	movf TempA,W
	xorlw 0xFF
	btfss STATUS,Z	
	goto MainLoop		; CMD error

; read/write WIF program name
	movf TempB,W
	call WInumToEEP
	movlw pcEEPinit
	call TransWt
	goto pgnDTread

NMread	movf TempA,W
	xorlw pcEEPread
	btfss STATUS,Z
	goto NMwrit

	movlw pcEEPread
	call TransWt
	movlw 4h
	movwf TempB
SendNm4	call RDeeprom
	call TransWt
	decfsz TempB
	goto SendNm4
	goto pgnDTread

; write name char
NMwrit	movf TempA,W
	xorlw pcEEPwrite
	btfss STATUS,Z
	goto NMend

	call Receive
	call WTeeprom		; write data
	movlw pcEEPwrite
	call TransWt		; ACK
	goto pgnDTread

; name 
NMend	movf TempA,W
	xorlw pcEEPend
	btfss STATUS,Z
	goto MainLoop		; CMD error

	movlw pcEEPend		; finished
	call TransWt
	goto MainLoop
;
; ------------------------
; PC READ FROM Memory Chip
; ------------------------
;
pcReadEEP	call BlankDisp		; clear display
	movlw rzMainCmdRDprg		; ACK ready
	call TransWt

	call Receive		; EEPROM block address for program start
	xorlw pcEEPaddr
	btfss STATUS,Z
	goto ACKwerr		; invalid

	call Receive
	movwf EEaddrH
	clrf EEaddrL
	movlw XFERbuff		; base address for EEPROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	call StartRD		; initiates EEP reading
	movlw pcEEPaddr
	call TransWt

readREQ	call Receive		; does PC want data
	xorlw pcEEPend
	btfss STATUS,Z
	goto doDatRead		; yes

	movlw pcEEPend		; no, ACK finished
	call TransWt
	bcf genBits2,n_ack		; no EEPROM ACK required - do EESTOP
	call EEread
	goto MainLoop

doDatRead	movf TempA,W
	xorlw pcEEPread
	btfss STATUS,Z
	goto ACKwerr		; invalid

	movlw pcEEPread
	call TransWt
	movlw d'4'			; 4 bytes
	movwf TempC

rXferLoop	call EEread
	call TransWt
	decfsz TempC
	goto rXferLoop
	goto readREQ
;
; -----------------------
; PC Write To Memory Chip
; -----------------------
;
pcWritEEP	call BlankDisp		; clear display
	movlw rzMainCmdWTprg		; ACK ready
	call TransWt

	call Receive		; EEPROM block address for program start
	xorlw pcEEPaddr
	btfss STATUS,Z
	goto ACKwerr		; received data error

	call Receive		; HI byte of block address for 32 bytes to come
	movwf EEaddrH
	clrf EEaddrL		; LO byte = 0
	movlw pcEEPaddr
	call TransWt		; ACK address received

doWRTloop	call Receive		; what next??
	xorlw pcEEPend
	btfss STATUS,Z
	goto doDatIni

	movlw pcEEPend		; no more data, ACK finished
	call TransWt
	goto MainLoop

doDatIni	movf TempA,W
	xorlw pcEEPinit
	btfss STATUS,Z
	goto doWrtDat

	movlw XFERbuff		; init base address for EEPROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	movlw pcEEPdata
	call TransWt
	goto doWRTloop

doWrtDat	movf TempA,W
	xorlw pcEEPdata
	btfss STATUS,Z
	goto izWrite		; write the block

get4Data	movlw d'4'			; write 4 bytes
	movwf TempC
BufDatLP1	call Receive
	movwf POSTINC0
	decfsz TempC
	goto BufDatLP1

	movlw pcEEPdata		; ACK data received
	call TransWt
	goto doWRTloop

izWrite	movf TempA,W
	xorlw pcEEPwrite
	btfss STATUS,Z
	goto MainLoop		; CMD error
;
; write the buffer data to EXT eeprom
;
	clrf FSR0L			; reset base address for EEPROM data buffer storage
	bcf genBits1,fgErase
	call wrtEEP32
	btfsc genBits2,EEerror
	goto ACKwerr

	movlw XFERbuff		; reset base address for EEPROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	movlw pcEEPwrite
	call TransWt
	goto doWRTloop

ACKwerr	call EEstop			; write error occurred
	call RstRS232
	movlw pcEEPerror
	call TransWt
	goto MainLoop
;
; PC - erase program on memory chip
;
pcErasEEP	movlw rzMainCmdDLprg
	call TransWt

	call Receive
	movf TempA,W
	xorlw pcEEPinit
	btfss STATUS,Z
	goto MainLoop		; CMD error

eeErazLP	movlw pcEEPwrite
	call TransWt

	call Receive
	xorlw pcEEPend
	btfss STATUS,Z
	goto erazMOR

	movlw pcEEPend
	call TransWt
	goto MainLoop

erazMOR	movf TempA,W
	xorlw pcEEPwrite
	btfss STATUS,Z
	goto MainLoop		; C<D error

	call Receive		; program number
	movwf TempB
	call Receive		; EDUC = 00, WIF = FF
	xorlw 0x00
	btfsc STATUS,Z
	goto erazED

	movf TempA,W
	xorlw 0xFF
	btfss STATUS,Z
	goto MainLoop		; CMD error

	movf TempB,W
	call WInumToEEP
	goto erazPGM

erazED	movf TempB,W
	call EDnumToEEP
erazPGM	call RDeepromNI
	iorlw prgNotValidMask
	call WTeeprom
	goto eeErazLP
;
; ----------------------------
; PROGRAM LOAD/SAVE ERROR LOOP
; ----------------------------
;
ErrorStart	bcf errorFlags,efgOnOff
	movf PCregL,W
	movwf tempPC
	movf MAreg,W
	movwf tempMA
	movf MBreg,W
	movwf tempMB
	movf ACreg,W
	movwf tempAC
	movf insLEDs,W
	movwf tempINS
	movf rfdeLEDs,W
	movwf tempRFDE
	clrf PCregL
	clrf PCregH
	clrf MAreg
	clrf MBreg
	clrf ACreg
	clrf insLEDs
	clrf rfdeLEDs
ETinit	bsf errorFlags,efgInitOff
ETimeLoop	clrf DelA			; about 1/4 sec
	movlw 0x05
	movwf DelB

ErrorLoop
	BANKSEL PIR0
ELwait	btfss PIR0,TMR0IF,BANKED
	goto ELwait

	bcf PIR0,TMR0IF,BANKED		; 128uS

	decfsz LEDupdate
	goto FlashError

	call ScanKeyLED		; keep updating display and test LA switch
	movf LEDupCnt,W
	movwf LEDupdate
	btfss errorFlags,efgERRdisplay
	goto ExitError		; LA pressed - return to normal

FlashError	decfsz DelA			; about 1/4 second delay
	goto ErrorLoop
	decfsz DelB
	goto ErrorLoop

	btfss errorFlags,efgInitOff	; small delay to hold display OFF
	goto fsOKerror

	bcf errorFlags,efgInitOff
	btfsc errorFlags,efgExit
	goto ExitError

fsOKerror	btfss errorFlags,efgOK		; Load/Save OK, flash PCregL[0] once
	goto pgmErr12

	btfsc errorFlags,efgOnOff
	goto fsOKend

	movlw 0x01
	movwf PCregL
	bsf errorFlags,efgOnOff
	goto ETimeLoop

fsOKend	clrf PCregL
	clrf PCregH
	bsf errorFlags,efgInitOff
	bsf errorFlags,efgExit
	goto ETinit			; after next OFF time - exit

pgmErr12	btfsc errorFlags,efgOnOff
	goto Err12off

	movlw 0x01
	btfss errorFlags,efgError01	; Load/Save Fail, flash PCregL[0] continuous
	movlw 0x81			; Load/Save EEPROM access fail, flash PCregL[7][0] continuous
	movwf PCregL
	bsf errorFlags,efgOnOff
	goto ETimeLoop

Err12off	clrf PCregL
	clrf PCregH
	bcf errorFlags,efgOnOff
	goto ETimeLoop


ExitError	clrf errorFlags
	movf tempPC,W
	movwf PCregL
	movf tempMA,W
	movwf MAreg
	movf tempMB,W
	movwf MBreg
	movf tempAC,W
	movwf ACreg
	movf tempINS,W
	movwf insLEDs
	movf tempRFDE,W
	movwf rfdeLEDs
	return
;
; --------------------------
; EXECUTE INSTRUCTION CYCLES
; --------------------------
;
CodeExec	btfss rfdeLEDs,Run
	goto MainLoop

	decfsz cycleCountL		; timing for FAST/SLOW mode
	goto MainLoop
	decfsz cycleCountH
	goto MainLoop

	if TestMode == 1
	  movlw 0x01		; FAST mode = 0x0001 48uS
	  movwf cycleCountH

	  movlw 0x01
	  movwf cycleCountL
	else
	  movlw 0x01		; FAST mode = 0x0001 48uS
	  btfss genBits1,fgSlowFast
	  movf SlowCalH,W		; SLOW mode = 0x28B0 500mS
	  movwf cycleCountH

	  movlw 0x01
	  btfss genBits1,fgSlowFast
	  movf SlowCalL,W
	  movwf cycleCountL
	endif

	btfsc rfdeLEDs,wiWIF		; user flag set for WIF mode
	goto WIFins
	goto NRMcode

	org 0x1300
;
; ---------------------------------------
; SUBROUTINE: GET BLANK PROGRAM NAME TEXT
; ---------------------------------------
;
BlankName	movlw HIGH BlankName
	movwf PCLATH
	bcf STATUS,C
	rlcf TempB,W
	addwf PCL
	retlw "P"
	retlw "r"
	retlw "o"
	retlw "g"
	retlw "r"
	retlw "a"
	retlw "m"
	retlw " "
	retlw 0xFF
;
; -----------------------------------
; SUBROUTINE: GET PROGRAM NUMBER TEXT
; -----------------------------------
;
ProgNumTxt	movlw HIGH ProgNumTxt
	movwf PCLATH
	bcf STATUS,C
	rlcf TempA,W
	andlw 0x0E
	addwf PCL
	retlw "1"
	retlw "2"
	retlw "3"
	retlw "4"
	retlw "5"
	retlw "6"
	retlw "7"
	retlw "8"
;
; ------------------------------------------------------------
; SUBROUTINE: IF WIF CODE HAS [RESULT SRC TGT] THEN n_ack=True
; ------------------------------------------------------------
;
codeRslt	bsf genBits2,n_ack
	movf wifADDR,W
	xorlw picADDR
	btfsc STATUS,Z
	return
	movf wifADDR,W
	xorlw picANDR
	btfsc STATUS,Z
	return
	movf wifADDR,W
	xorlw picIORR
	btfsc STATUS,Z
	return
	movf wifADDR,W
	xorlw picSUBR
	btfsc STATUS,Z
	return
	movf wifADDR,W
	xorlw picXORR
	btfss STATUS,Z
	bcf genBits2,n_ack
	return
;
; ----------------------------
; SUBROUTINE: MATCH TGT TO SRC
; ----------------------------
; SETB CLRB ADDR ANDR IORR SUBR XORR
matTGTtoSRC	movlw maskTGT
	andwf wiCodeL
	btfsc wiCodeH,1
	bsf wiCodeL,7
	btfsc wiCodeH,0
	bsf wiCodeL,6
	return

	org 0x1400

NRMcode	movlw HIGH NRMcode
	movwf PCLATH
	movf execFunction,W
	addwf PCL
	goto MainLoop
	goto ExecDep
	goto ExecExam
	goto ExecFetch
	goto ExecDefer
	goto ExecExec
;
; ---------------
; EXECUTE DEPOSIT
; ---------------
;
ExecDep	movlw HIGH ExecDep
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeDep0		; Cycle 0
	goto ExeDep1		; Cycle 1
	goto ExeDep29		; Cycle 2
	goto ExeDep29		; Cycle 3
	goto ExeDep29		; Cycle 4
	goto ExeDep29		; Cycle 5
	goto ExeDep29		; Cycle 6
	goto ExeDep29		; Cycle 7
	goto ExeDep29		; Cycle 8
	goto ExeDep29		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeDep12		; Cycle 12
	goto ExeDep13		; Cycle 13
	goto ExeDep1421		; Cycle 14
	goto ExeDep1421		; Cycle 15
	goto ExeDep1421		; Cycle 16
	goto ExeDep1421		; Cycle 17
	goto ExeDep1421		; Cycle 18
	goto ExeDep1421		; Cycle 19
	goto ExeDep1421		; Cycle 20
	goto ExeDep1421		; Cycle 21
	goto MainLoop		; Cycle 22
	goto ExeDep23		; Cycle 23
;
; ---------------
; EXECUTE EXAMINE
; ---------------
;
ExecExam	movlw HIGH ExecExam
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeExam0		; Cycle 0
	goto ExeExam1		; Cycle 1
	goto ExeExam29		; Cycle 2
	goto ExeExam29		; Cycle 3
	goto ExeExam29		; Cycle 4
	goto ExeExam29		; Cycle 5
	goto ExeExam29		; Cycle 6
	goto ExeExam29		; Cycle 7
	goto ExeExam29		; Cycle 8
	goto ExeExam29		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto MainLoop		; Cycle 12
	goto ExeExam13		; Cycle 13
	goto ExeExam1421		; Cycle 14
	goto ExeExam1421		; Cycle 15
	goto ExeExam1421		; Cycle 16
	goto ExeExam1421		; Cycle 17
	goto ExeExam1421		; Cycle 18
	goto ExeExam1421		; Cycle 19
	goto ExeExam1421		; Cycle 20
	goto ExeExam1421		; Cycle 21
	goto MainLoop		; Cycle 22
	goto ExeExam23		; Cycle 23

	org 0x1500
;
; ---------------
; EXECUTE FETCH
; ---------------
;
ExecFetch	movlw HIGH ExecFetch
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeFet0		; Cycle 0
	goto ExeFet1		; Cycle 1
	goto ExeFet29		; Cycle 2
	goto ExeFet29		; Cycle 3
	goto ExeFet29		; Cycle 4
	goto ExeFet29		; Cycle 5
	goto ExeFet29		; Cycle 6
	goto ExeFet29		; Cycle 7
	goto ExeFet29		; Cycle 8
	goto ExeFet29		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeFet12		; Cycle 12
	goto MainLoop		; Cycle 13
	goto ExeFet1420		; Cycle 14
	goto ExeFet1420		; Cycle 15
	goto ExeFet1420		; Cycle 16
	goto ExeFet1420		; Cycle 17
	goto ExeFet1420		; Cycle 18
	goto ExeFet1420		; Cycle 19
	goto ExeFet1420		; Cycle 20
	goto ExeFet21		; Cycle 21
	goto MainLoop		; Cycle 22
	goto ExeFet23		; Cycle 23
;
; -------------
; EXECUTE DEFER
; -------------
;
ExecDefer	movlw HIGH ExecDefer
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeDef0		; Cycle 0
	goto ExeDef1		; Cycle 1
	goto MainLoop		; Cycle 2
	goto MainLoop		; Cycle 3
	goto MainLoop		; Cycle 4
	goto MainLoop		; Cycle 5
	goto MainLoop		; Cycle 6
	goto MainLoop		; Cycle 7
	goto MainLoop		; Cycle 8
	goto MainLoop		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeDef12		; Cycle 12
	goto ExeDef13		; Cycle 13
	goto ExeDef1421		; Cycle 14
	goto ExeDef1421		; Cycle 15
	goto ExeDef1421		; Cycle 16
	goto ExeDef1421		; Cycle 17
	goto ExeDef1421		; Cycle 18
	goto ExeDef1421		; Cycle 19
	goto ExeDef1421		; Cycle 20
	goto ExeDef1421		; Cycle 21
	goto MainLoop		; Cycle 22
	goto ExeDef23		; Cycle 23

	org 0x1600
;
; ---------------
; EXECUTE EXECUTE
; ---------------
;
ExecExec	bsf fgsExecDefer,Execute
	movlw b'11000000'		; is instruction < 6?
	cpfslt INSreg
	goto IOorAUG		; no
;
; 0 - 5 = memory reference instruction
;
ExecMrfIns	movlw HIGH ExecExec
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeMrf0		; Cycle 0
	goto ExeMrf1		; Cycle 1
	goto ExeMrf29		; Cycle 2
	goto ExeMrf29		; Cycle 3
	goto ExeMrf29		; Cycle 4
	goto ExeMrf29		; Cycle 5
	goto ExeMrf29		; Cycle 6
	goto ExeMrf29		; Cycle 7
	goto ExeMrf29		; Cycle 8
	goto ExeMrf29		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeMrf12		; Cycle 12
	goto ExeMrf13		; Cycle 13
	goto ExeMrf1421		; Cycle 14
	goto ExeMrf1421		; Cycle 15
	goto ExeMrf1421		; Cycle 16
	goto ExeMrf1421		; Cycle 17
	goto ExeMrf1421		; Cycle 18
	goto ExeMrf1421		; Cycle 19
	goto ExeMrf1421		; Cycle 20
	goto ExeMrf1421		; Cycle 21
	goto MainLoop		; Cycle 22
	goto ExeMrf23		; Cycle 23
;
; IO or Augment instruction?
;
IOorAUG	movf INSreg,W
	andlw b'11100000'
	xorlw b'11000000'
	btfss STATUS,Z
	goto doAUGins
;
; 6 = IO instruction
;
ExecIO	btfsc INSreg,4		; test bit 4 to see what type of Augment instruction to execute
	goto ExecIOoEX
	
	movlw HIGH ExecIO
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeIO0i		; Cycle 0
	goto ExeIO1i		; Cycle 1
	goto ExeIO29i		; Cycle 2
	goto ExeIO29i		; Cycle 3
	goto ExeIO29i		; Cycle 4
	goto ExeIO29i		; Cycle 5
	goto ExeIO29i		; Cycle 6
	goto ExeIO29i		; Cycle 7
	goto ExeIO29i		; Cycle 8
	goto ExeIO29i		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeIO12i		; Cycle 12
	goto ExeIO13i		; Cycle 13
	goto ExeIO14i		; Cycle 14
	goto ExeIO1521i		; Cycle 15
	goto ExeIO1521i		; Cycle 16
	goto ExeIO1521i		; Cycle 17
	goto ExeIO1521i		; Cycle 18
	goto ExeIO1521i		; Cycle 19
	goto ExeIO1521i		; Cycle 20
	goto ExeIO1521i		; Cycle 21
	goto MainLoop		; Cycle 22
	goto EndEXEC		; Cycle 23 finished

	org 0x1700

ExecIOoEX	movlw HIGH ExecIOoEX
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeIO0o		; Cycle 0
	goto ExeIO1o		; Cycle 1
	goto ExeIO2o		; Cycle 2
	goto ExeIO39o		; Cycle 3
	goto ExeIO39o		; Cycle 4
	goto ExeIO39o		; Cycle 5
	goto ExeIO39o		; Cycle 6
	goto ExeIO39o		; Cycle 7
	goto ExeIO39o		; Cycle 8
	goto ExeIO39o		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeIO12o		; Cycle 12
	goto ExeIO13o		; Cycle 13
	goto ExeIO14o		; Cycle 14
	goto ExeIO1421o		; Cycle 15
	goto ExeIO1421o		; Cycle 16
	goto ExeIO1421o		; Cycle 17
	goto ExeIO1421o		; Cycle 18
	goto ExeIO1421o		; Cycle 19
	goto ExeIO1421o		; Cycle 20
	goto ExeIO1421o		; Cycle 21
	goto MainLoop		; Cycle 22
	goto EndEXEC		; Cycle 23 finished
;
; 7 = Augment instruction
;
doAUGins	btfsc INSreg,4		; test bit 4 to see what type of Augment instruction to execute
	goto inExeAugH
	
	movlw HIGH doAUGins
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeAugL0		; Cycle 0
	goto ExeAugL1		; Cycle 1
	goto ExeAugL29		; Cycle 2
	goto ExeAugL29		; Cycle 3
	goto ExeAugL29		; Cycle 4
	goto ExeAugL29		; Cycle 5
	goto ExeAugL29		; Cycle 6
	goto ExeAugL29		; Cycle 7
	goto ExeAugL29		; Cycle 8
	goto ExeAugL29		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeAugL12		; Cycle 12
	goto MainLoop		; Cycle 13
	goto MainLoop		; Cycle 14
	goto MainLoop		; Cycle 15
	goto MainLoop		; Cycle 16
	goto MainLoop		; Cycle 17
	goto MainLoop		; Cycle 18
	goto MainLoop		; Cycle 19
	goto MainLoop		; Cycle 20
	goto MainLoop		; Cycle 21
	goto MainLoop		; Cycle 22
	goto EndEXEC		; Cycle 23 finished

	org 0x1800

inExeAugH	movlw HIGH inExeAugH
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto ExeAugH0		; Cycle 0
	goto MainLoop		; Cycle 1
	goto MainLoop		; Cycle 2
	goto MainLoop		; Cycle 3
	goto MainLoop		; Cycle 4
	goto MainLoop		; Cycle 5
	goto MainLoop		; Cycle 6
	goto MainLoop		; Cycle 7
	goto MainLoop		; Cycle 8
	goto MainLoop		; Cycle 9
	goto MainLoop		; Cycle 10
	goto MainLoop		; Cycle 11
	goto ExeAugH12		; Cycle 12
	goto ExeAugH13		; Cycle 13
	goto ExeAugH1421		; Cycle 14
	goto ExeAugH1421		; Cycle 15
	goto ExeAugH1421		; Cycle 16
	goto ExeAugH1421		; Cycle 17
	goto ExeAugH1421		; Cycle 18
	goto ExeAugH1421		; Cycle 19
	goto ExeAugH1421		; Cycle 20
	goto ExeAugH1421		; Cycle 21
	goto MainLoop		; Cycle 22
	goto EndEXEC		; Cycle 23 finished
;
; --------------
; DEPOSIT CYCLES
; --------------
;
ExeDep0	movf rfdeLEDs,W		; Cycle 0
	andlw maskRFDEoff
	iorlw maskSetRun		; FETCH OFF RUN ON
	movwf rfdeLEDs

	clrf MAreg			; Cycle 0
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeDep1	movf PCregL,W		; Cycle 1
	movwf ROLregA		; shift PC -> MA
	addlw 0x01
	movwf ROLregB		; shift PC + 1 -> PC

	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeDep29	rrcf ROLregA		; Cycle 2-9
	rrcf MAreg
	rrcf ROLregB
	rrcf PCregL
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeDep29

ExeDep12	btfss genBits1,fgSlowFast	; Cycle 12
	goto MainLoop

ExeDep13	movf binSwitch,W		; Cycle 13
	movwf MBreg			; switches -> MB
	movlw MEMbuff
	movwf FSR0H			; MB -> Memory[MA] - parallel load
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	movf MBreg,W
	movwf INDF0
	movf MBreg			; MB reloads during cycles 14-21
	movwf ROLregA		; shift MB -> MB
	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeDep1421	rrcf ROLregA		; Cycle 14-21
	rrcf MBreg
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeDep1421

ExeDep23	call ClearRUN		; Cycle 23
	goto MainLoop
;
; --------------
; EXAMINE CYCLES
; --------------
;
ExeExam0	movf rfdeLEDs,W		; cycle 0
	andlw maskRFDEoff
	iorlw maskSetRun		; FETCH OFF RUN ON
	movwf rfdeLEDs
	bsf fgsExecDefer,Run

	clrf MAreg			; Cycle 0
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeExam1	movf PCregL,W		; Cycle 1
	movwf ROLregA		; shift PC -> MA
	addlw 0x01
	movwf ROLregB		; shift PC + 1 -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeExam29	rrcf ROLregA		; Cycle 2-9
	rrcf MAreg
	rrcf ROLregB
	rrcf PCregL
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeExam29

ExeExam13	movlw MEMbuff		; Cycle 13 - Memory[MA] -> MB
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	movwf ROLregA
	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeExam1421	rrcf ROLregA		; Cycle 14-21
	rrcf MBreg			; shift Memory[MA] -> MB
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeExam1421

ExeExam23	call ClearRUN		; Cycle 23
	goto MainLoop
;
; ------------
; FETCH CYCLES
; ------------
;
ExeFet0	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetRunFetch		; RUN ON, FETCH ON
	movwf rfdeLEDs
	bsf fgsExecDefer,Run

	clrf MAreg			; Cycle 0
	clrf INSreg
	bsf fgsExecDefer,Fetch
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeFet1	movf PCregL,W		; Cycle 1
	movwf ROLregA		; shift PC -> MA
	addlw 0x01
	movwf ROLregB		; shift PC + 1 -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeFet29	rrcf ROLregA		; Cycle 2-9
	rrcf MAreg
	rrcf ROLregB
	rrcf PCregL	
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeFet29

ExeFet12	movlw MEMbuff		; Cycle 12
	movwf FSR0H			; Memory[MA] -> MB
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	movwf ROLregA		; shift Memory[MA] -> MB
	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 7h
	movwf TempC

ExeFet1420	rrcf ROLregA		; Cycle 14-20
	rrcf MBreg
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeFet1420

ExeFet21	rrcf ROLregA		; Cycle 21
	rrcf MBreg,W
	movwf INSreg
	movwf MBreg
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeFet23	bcf genBits1,fgDeferSet		; Cycle 23, DEFER set ONLY IF - bit 4 of instruction set AND it is a memory reference instruction (bits [765] = 0 - 5) Page 46
	movf MBreg,W
	andlw b'11100000'
	movwf TempD

	btfss MBreg,4		; bit 4 set?
	goto noDefer		; no

	movlw b'10100000'
	cpfsgt TempD
	bsf genBits1,fgDeferSet		; < 6 = Defer instruction

noDefer	btfss rfdeLEDs,PageZero		; page zero mode operating?
	goto izMRinst

; (bUsePageZero = True) and (MBreg and $08 = $00) and ((Ins < 4) or ((Ins = 4) and ((MBreg and $10) = $10)) { E = F.T23.MB3=0.(ANDTADISZDCAJMS.MB4) }
doPageZero	btfsc MBreg,3		; MBreg and $08 = 0?
	goto izMRinst		; no

	movlw b'10100000'		; is instrunction <= 4?
	cpfslt TempD
	goto izMRinst		; no

	movlw b'01100000'
	cpfsgt TempD		; is instrunction > 3?
	goto doZeroMode		; no, must be < 4

	btfss MBreg,4		; instruction = 4, MBreg and $10 = $10?
	goto izMRinst		; no

doZeroMode	movlw 0x0F		; do Page Zero mode  = mask upper address nibble
	andwf MBreg
	goto exitF23

izMRinst	movlw b'11000000'		; if a mem ref instruction, MB 7-4 = MA 7-4
	cpfslt TempD
	goto exitF23		; > 5, no

	movlw 0x0F
	andwf MBreg
	movf MAreg,W
	andlw 0xF0
	iorwf MBreg

;	movf MAreg,W		; iMB = (iMA and $F0) + (iMB and $0F)
;	andlw 0xF0
;	movwf TempC
;	movf MBreg,W
;	andlw 0x0F
;	iorwf TempC,W
;	movwf MBreg
	goto exitF23

exitF23	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetRun		; FETCH OFF RUN STAYS ON
	movwf rfdeLEDs

	movlw maskSetDefer		; DEFER ON?
	btfss genBits1,fgDeferSet
	movlw maskSetExecute		; EXECUTE ON?
	iorwf rfdeLEDs

	movlw initCycleIndex
	movwf execIndex
	movlw funcDefer		; defer or execute next
	btfss genBits1,fgDeferSet
	movlw funcExecute
	movwf execFunction
	goto MainLoop
;
; ------------
; DEFER CYCLES
; ------------
;
ExeDef0	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetRunDefer		; RUN ON DEFER ON
	movwf rfdeLEDs
	bsf fgsExecDefer,Defer

	clrf MAreg			; Cycle 0
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeDef1	movf MBreg,W		; Cycle 1 - MB -> MA
	movwf MAreg
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeDef12	movlw MEMbuff		; Cycle 12
	movwf FSR0H			; Memory[MA] -> MB
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	movwf ROLregA		; shift Memory[MA] -> MB
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeDef13	btfss genBits1,fgSlowFast	; Cycle 13
	goto MainLoop

	movlw 8h
	movwf TempC

ExeDef1421	rrcf ROLregA		; Cycle 14 - 21
	rrcf MBreg
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeDef1421
	
ExeDef23	movlw b'11101111'		; Cycle 23
	andwf INSreg		; clear defer bit - upsets instruction jump tables
	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetRunExec		; DEFER OFF RUN STAYS ON, EXECUTE ON
	movwf rfdeLEDs
	movlw initCycleIndex
	movwf execIndex
	movlw funcExecute		; execute next
	movwf execFunction
	goto MainLoop
;
; ---------------------------------
; EXECUTE CYCLES - Memory Reference
; ---------------------------------
;
ExeMrf0	swapf INSreg,W		; Cycle 0
	movwf TempB
	call getOPRLED
	movwf insLEDs		; set instruction LED

	clrf MAreg			; Cycle 0
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeMrf1	movf MBreg,W		; Cycle 1
	movwf MAreg			; MB -> MA
	swapf INSreg,W
	movwf TempB
	goto MRdecode1		; jump to instruction to start executing Cycle 1

mriAND1	movlw MEMbuff		; Cycle 1
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	movwf ROLregA		; Memory[MA] -> MB
	andwf ACreg,W		; Memory[MA] and AC -> AC
	movwf ROLregB
	goto ExeMrfEX1

mriTAD1	movlw MEMbuff		; Cycle 1
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	movwf ROLregA		; Memory[MA] -> MB
	addwf ACreg,W		; Memory[MA] + AC -> AC
	movwf ROLregB
	goto ExeMrfEX1

mriISZ1	movlw MEMbuff		; Cycle 1
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	movf INDF0,W
	addlw 0x01
	movwf ROLregA		; Memory[MA] + 1 -> MB
	goto ExeMrfEX1

mriDCA1	clrf ROLregA		; Cycle 1 - 0 -> AC
	movf ACreg,W
	movwf ROLregB		; AC -> MB
	movwf ROLregC		; AC -> Memory[MA]
	goto ExeMrfEX1

mriJMS1	movf PCregL,W		; Cycle 1
	movwf ROLregA		; PC -> MB
	movwf ROLregB		; PC -> Memory[MA]
	movwf ROLregC		; PC -> PC

mriJMP1				; Cycle 1 - Nothing
mriIOT1				; Cycle 1 - Nothing
mriAUG1				; Cycle 1 - Nothing
ExeMrfEX1	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeMrf29	swapf INSreg,W
	movwf TempB
	goto MRdecode29		; jump to instruction to start executing Cycle 2-9

mriAND29	rrcf ROLregA		; Cycle 2 - 9
	rrcf MBreg			; Memory[MA] -> MB
	rrcf ROLregB
	rrcf ACreg			; Memory[MA] and AC -> AC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriAND29
	goto ExeMrf12

mriTAD29	rrcf ROLregA		; Cycle 2 - 9
	rrcf MBreg			; Memory[MA] -> MB
	rrcf ROLregB
	rrcf ACreg			; Memory[MA] + AC -> AC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriTAD29
	goto ExeMrf12

mriISZ29	rrcf ROLregA		; Cycle 2 - 9
	rrcf MBreg			; Memory[MA] + 1 -> MB
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriISZ29
	goto ExeMrf12

mriDCA29	rrcf ROLregA		; Cycle 2 - 9
	rrcf ACreg			; 0 -> AC
	rrcf ROLregB
	rrcf MBreg			; AC -> MB
	movlw MEMbuff
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	rrcf ROLregC
	rrcf INDF0			; AC -> Memory[MA]
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriDCA29
	goto ExeMrf12

mriJMS29	rrcf ROLregA		; Cycle 2 - 9
	rrcf MBreg			; PC -> MB
	movlw MEMbuff
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	rrcf ROLregB
	rrcf INDF0			; PC -> Memory[MA]
	rrcf ROLregC
	rrcf PCregL			; PC -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriJMS29


mriJMP29				; Cycle 2-9 - Nothing
mriIOT29				; Cycle 2-9 - Nothing
mriAUG29				; Cycle 2-9 - Nothing
	btfss genBits1,fgSlowFast
	goto MainLoop

; ---------------------------------------------------

ExeMrf12	btfss genBits1,fgSlowFast	; Cycle 12
	goto MainLoop

ExeMrf13	swapf INSreg,W
	movwf TempB
	goto MRdecode13		; jump to instruction to start executing Cycle 13

mriAND13	goto ExeMrfEX3		; Cycle 13

mriTAD13	goto ExeMrfEX3		; Cycle 13

mriISZ13	movf PCregL,W		; Cycle 13, PC -> PC
	movf MBreg			; if MB = 0, PC + 1 -> PC
	btfsc STATUS,Z
	addlw 1h
	movwf ROLregA
	movf MBreg,W
	movwf ROLregB		; MB -> Memory[MA]
	goto ExeMrfEX3

mriDCA13	goto ExeMrfEX3		; Cycle 13

mriJMS13	incf MAreg,W		; Cycle 13
	movwf ROLregA		; MA + 1 -> PC
	goto ExeMrfEX3

mriJMP13	movf MAreg,W		; Cycle 13
	movwf ROLregA		; MA -> PC

mriIOT13				; Cycle 13

mriAUG13				; Cycle 13
ExeMrfEX3	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

; ---------------------------------------------------

ExeMrf1421	swapf INSreg,W		; Cycle 14-21
	movwf TempB
	goto MRdecod1421		; jump to instruction to start executing Cycle 14-21

mriAND1421	btfss genBits1,fgSlowFast	; Cycle 14-21
	goto MainLoop
	goto ExeMrf23

mriTAD1421	btfss genBits1,fgSlowFast	; Cycle 14-21
	goto MainLoop
	goto ExeMrf23

mriISZ1421	movlw MEMbuff		; Cycle 14-21
	movwf FSR0H
	movf MAreg,W
	movwf FSR0L
	rrcf ROLregB
	rrcf INDF0			; MB -> Memory[MA]
	rrcf ROLregA
	rrcf PCregL			; PC + 1 or MA + 1 -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriISZ1421
	goto ExeMrf23

mriDCA1421	btfss genBits1,fgSlowFast	; Cycle 14-21
	goto MainLoop
	goto ExeMrf23

mriJMS1421	rrcf ROLregA		; Cycle 14-21
	rrcf PCregL			; PC + 1 or MA + 1 -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriJMS1421
	goto ExeMrf23

mriJMP1421	rrcf ROLregA		; Cycle 14-21
	rrcf PCregL			; PC + 1 or MA + 1 -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto mriJMP1421

mriIOT1421				; Cycle 14-21

mriAUG1421				; Cycle 14-21 (Follow on)

ExeMrf23				; Cycle 23 - follow on the END ececution (Follow On)

; ------------------------------------------------------

EndEXEC	btfsc genBits1,fgHALT		; Cycle 23 stop execution if HALT set
	goto doHALT

	btfss genBits1,fgSinCon		; stop execution if SINGLE set
	goto doHALT

	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetRunFetch		; EXEC OFF, RUN STAYS ON, FTECH ON
	goto FinExndEx

doHALT	bcf genBits1,fgHALT
	clrf insLEDs		; all instruction LEDs off
	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetFetch		; EXEC OFF, RUN OFF, FTECH ON

FinExndEx	movwf rfdeLEDs
	movlw initCycleIndex
	movwf execIndex
	movlw funcFetch		; execute fetch next
	movwf execFunction
	btfss genBits1,fgSlowFast
	clrf insLEDs		; all instruction LEDs off
	bcf genBits1,fgSHdata
	goto MainLoop
;
; ----------------------------------------------
; EXECUTE CYCLES - Augment Bit 4 LO instructions
; ----------------------------------------------
;
ExeAugL0	swapf INSreg,W		; Cycle 0
	movwf TempB
	call getOPRLED
	movwf insLEDs		; set instruction LED

	clrf MAreg			; Cycle 0
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeAugL1	movf ACreg,W		; Cycle 1
	movwf ROLregA

	btfss INSreg,3
	goto AugCMA

	clrf ROLregA		; CLA

AugCMA	btfss INSreg,2
	goto AugRAL

	movlw 0xFF			; CMA
	xorwf ROLregA

AugRAL	btfsc INSreg,1
	rlncf ROLregA		; RAL

AugIAC	btfss INSreg,0
	goto tstAugML

	incf ROLregA		; IAC

tstAugML	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeAugL29	rrcf ROLregA		; cycle 2-9
	rrcf ACreg			; Result -> AC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeAugL29

ExeAugL12	btfss genBits1,fgSlowFast	; Cycle 12
	goto MainLoop
	goto EndEXEC
;
; ----------------------------------------------
; EXECUTE CYCLES - Augment Bit 4 HI instructions
; ----------------------------------------------
;
ExeAugH0	swapf INSreg,W		; Cycle 0
	movwf TempB
	call getOPRLED
	movwf insLEDs		; set instruction LED

	clrf MAreg			; Cycle 0
	btfss genBits1,fgSlowFast
	goto MainLoop

ExeAugH12	btfss genBits1,fgSlowFast	; Cycle 12
	goto MainLoop

ExeAugH13	movlw 0h			; Cycle 13
	btfss INSreg,3
	goto AugSMA

	movf ACreg			; SZA
	btfsc STATUS,Z		; #### this line had ,C - 120919
	movlw 1h			; if AC = 0, (set carry)

AugSMA	btfss INSreg,2
	goto AugRAR

	btfsc ACreg,7		; SMA
	movlw 1h			; if AC = -ve, (set carry)

AugRAR	btfsc INSreg,1
	rrncf ACreg			; RAR

AugHLT	btfsc INSreg,0		; SZA and or SMA
	bsf genBits1,fgHALT

	addwf PCregL,W		; PC -> PC, or PC + 1 -> PC
	movwf ROLregA
	btfss genBits1,fgSlowFast
	goto MainLoop

	movlw 8h
	movwf TempC

ExeAugH1421	rrcf ROLregA		; cycle 14-21
	rrcf PCregL			; Result -> PC
	btfss genBits1,fgSlowFast
	goto MainLoop

	decfsz TempC
	goto ExeAugH1421

	btfss genBits1,fgSlowFast
	goto MainLoop
	goto EndEXEC
;
; --------------------------------
; EXECUTE CYCLES - IO instructions
; --------------------------------
; IO IN
ExeIO0i	swapf INSreg,W		; Cycle 0
	movwf TempB
	call getOPRLED
	movwf insLEDs		; set instruction LED

	clrf MAreg
	goto MainLoop

ExeIO1i	btfsc INSreg,3		; Cycle 1
	goto devIN1

	selIOin0			; select CLK/RstOUT DEV 0 IN
	goto MainLoop

devIN1	
	selIOin1			; select CLK/RstOUT DEV 1 IN
	goto MainLoop

ExeIO29i	btfss INSreg,1		; Cycle 2-9, KRS?
	goto MainLoop		; no

	clrIOclk			; Cycle 2 - 9, CLOCK = LO
	call CKdelay		; wait for stabilize

	btfsc INSreg,3
	goto SHdevIN1

	bcf STATUS,C
	skipIN0DAlo			; read DEV 0 INPUT
	bsf STATUS,C
	goto ExeIO29EXa

SHdevIN1	bcf STATUS,C
	skipIN1DAlo			; read DEV 1 INPUT
	bsf STATUS,C
ExeIO29EXa	rrcf ACreg
	setIOclk			; CLOCK = HI
	goto MainLoop

ExeIO12i	goto MainLoop

ExeIO13i	bcf genBits1,fgSHdata		; Cycle 13
	movlw 1h			; preset PC + 1 -> PC
	btfss INSreg,0
	goto testRKF

	btfsc INSreg,3		; SKF
	goto SKFin1

	skipIN0RIlo			; test if DEV 0 INPUT reset = LO
	movlw 0h			; no, PC -> PC
	goto modPC13i

SKFin1
	skipIN1RIlo			; test if DEV 1 INPUT reset = LO
	movlw 0h			; no, PC -> PC

modPC13i	addwf PCregL,W
	movwf ROLregA
	bsf genBits1,fgSHdata		; PC data to shift during cycles 14-21

; follow on the RKF

testRKF	btfss INSreg,2		; is RKF?
	goto MainLoop		; no

	clrIOrst			; DEV X RESET = LO
	goto MainLoop

ExeIO14i				; cycle 14 RKF, IO select already set
	setIOrst			; DEV X RESET = HI

ExeIO1521i	btfss genBits1,fgSHdata		; Cycle 15-21
	goto MainLoop

	rrcf ROLregA
	rrcf PCregL			; PC or PC + 1 -> PC
	goto MainLoop
;
; --------------------------------
; EXECUTE CYCLES - IO instructions
; --------------------------------
; IO OUT
ExeIO0o	swapf INSreg,W		; Cycle 0
	movwf TempB
	call getOPRLED
	movwf insLEDs		; set instruction LED

	clrf MAreg
	goto MainLoop

ExeIO1o	btfsc INSreg,3		; cycle 1
	goto devOUT1

	selIOout0			; select CLK/RstOUT DEV 0 OUT
	goto InsLDS

devOUT1
	selIOout1			; select CLK/RstOUT DEV 1 OUT

InsLDS	btfss rfdeLEDs,ResetT1T13	; if INS is LDS RDF and Reset on T1 = True
	goto shDevOs

	btfss INSreg,2
	goto shDevOs

	clrIOrst

shDevOs	movf ACreg,W		; AC -> DEV x OUT
	movwf ROLregA
	clrf ROLregB		; 0 -> AC
	goto MainLoop

ExeIO2o				; Cycle 2 - May not be required, but reset IO flag anyway to make sure it is HI
	setIOrst

; follow on the shift data cycles

ExeIO39o	btfss INSreg,1		; Cycle (2) 3-9
	goto MainLoop

	rrcf ROLregA
	btfsc INSreg,3
	goto dev1sh

	btfsc STATUS,C
	goto Dev0hi

	clrDataOut0			; dev 0 out = LO
	goto ExeIO29ck

Dev0hi	setDataOut0			; dev 0 out = HI
	goto ExeIO29ck
	
dev1sh	btfsc STATUS,C
	goto Dev1hi

	clrDataOut1			; dev 1 out = LO
	goto ExeIO29ck

Dev1hi	setDataOut1			; dev 1 out = HI
	goto ExeIO29ck

ExeIO29ck	call CKdelay
	clrIOclk

	rrcf ROLregB		; shift regB -> AC
	rrcf ACreg

	call CKdelay
	setIOclk
	goto MainLoop

ExeIO12o	goto MainLoop		; Cycle 12

ExeIO13o	bcf genBits1,fgSHdata		; Cycle 13
	movlw 1h			; preset PC + 1 -> PC (output x flag = LO)
	btfss INSreg,0
	goto rstT13

	btfsc INSreg,3		; SDF
	goto SDFout1

	skipOUT0RIlo
	movlw 0h
	goto modPC13o

SDFout1
	skipOUT1RIlo
	movlw 0h

modPC13o	addwf PCregL,W		; output x flag = HI, PC -> PC
	movwf ROLregA
	bsf genBits1,fgSHdata

rstT13	btfsc rfdeLEDs,ResetT1T13	; if INS is LDS RDF and Reset on T13 = True
	goto MainLoop

	btfss INSreg,2
	goto MainLoop

	clrIOrst
	goto MainLoop

ExeIO14o				; Cycle 14 - May not be required, but reset IO flag anyway to make sure it is HI
	setIOrst

ExeIO1421o	btfss genBits1,fgSHdata
	goto MainLoop

	rrcf ROLregA		; Cycle 14-21
	rrcf PCregL			; PC or PC + 1 -> PC
	goto MainLoop
;
; -------------------------------
; SUBROUTINE: GET SOURCE REGISTER
; -------------------------------
;
GetSRCreg	rrncf wiCodeH,W		; 1000 00ss
	movwf TempD
	rrncf TempD,W
	goto InGetREG
;
; -------------------------------
; SUBROUTINE: GET TARGET REGISTER
; -------------------------------
;
GetTGTreg	movf wiCodeL,W
InGetREG	movwf TempD

	btfsc TempD,7		; tt00 0000
	goto TGT1x

	btfsc TempD,6
	goto TGT01

TGT00	movf ACreg,W		; AC
	return

TGT01	movlw RAMbuff		;  R = aRAM[iCode and RAMPAGEsize]
	movwf FSR0H
	movf wiCodeL,W
	andlw wiRAMPAGEmax
	movwf FSR0L
	movlw 0x20
	btfsc statusWIF,wifRP
	addwf FSR0L			; RAM Page 1
	movf INDF0,W
	return

TGT1x	btfsc TempD,6
	goto TGT11

TGT10	movlw RAMbuff		;  I = aRAM[aRAM[0]]
	movwf FSR0H
	clrf FSR0L
	movf INDF0,W
	andlw wiRAMmax		; limit INDR to max RAM size
	movwf FSR0L
	movf INDF0,W
	return

TGT11	movf PCregL,W		; P = PC low 8 bits
	return
;
; --------------------------
; SUBROUTINE: PUT TARGET REG
; --------------------------
;
PutTGTreg	btfsc wiCodeL,7		; tt00 0000
	goto ptTGT1x

	btfsc wiCodeL,6
	goto ptTGT01

	movf MBreg,W		; AC
	movwf ACreg
	return

ptTGT01	movlw RAMbuff		;  R = aRAM[iCode and RAMPAGEsize]
	movwf FSR0H
	movf wiCodeL,W
	andlw wiRAMPAGEmax
	movwf FSR0L
	movlw 0x20
	btfsc statusWIF,wifRP
	addwf FSR0L			; RAM Page 1
	movf MBreg,W
	movwf INDF0
	return

ptTGT1x	btfsc wiCodeL,6
	goto ptTGT11

	movlw RAMbuff		;  I = aRAM[aRAM[0]]
	movwf FSR0H
	clrf FSR0L
	movf INDF0,W
	andlw wiRAMmax		; limit INDR to max RAM size
	movwf FSR0L
	movf MBreg,W
	movwf INDF0
	return

ptTGT11	movf ROMaddrH,W		; iPC = ROMaddrH + ((iMB + 1) and $FF);
	movwf PCregH
	incf MBreg,W
	movwf PCregL
	bsf genBitsWIF,wifPCmod		; no increment PC after instruction completes, or addskip
	return

	org 0x2000
;
; jump table for WIF instructions
;
DecodeWIF	btfsc wifTblMain,7
	goto DecodeWIF2

	movlw HIGH DecodeWIF
	movwf PCLATH
	movf wifTblMain,W
	addwf PCL
	goto idwINS00	; idwNOP		0
	goto idwINS00	; idwADDVA		1
	goto idwINS00	; idwANDVA		2
	goto idwINS00	; idwIORVA		3
	goto idwINS00	; idwMOVVA		4
	goto idwINS00	; idwSUBVA		5
	goto idwINS00	; idwXORVA		6
	goto idwINS00	; idwSKIPA		7
	goto idwINS00	; idwRDSWA		8
	goto idwINS00	; idwDAAC		9
	goto idwINS00	; idwMULTA		10
	goto idwINS00	; idwRANDA		11
	goto wiEndExec	; NO INS		12
	goto wiEndExec	; NO INS		13
	goto wiEndExec	; NO INS		14
	goto wiEndExec	; NO INS		15
	goto idwINS01	; idwADDR		16
	goto idwINS01	; idwANDR		17
	goto idwINS01	; idwCLRR		18
	goto idwINS01	; idwDECR		19
	goto idwINS01	; idwDECSZ		20
	goto idwINS01	; idwINCR		21
	goto idwINS01	; idwINCSZ		22
	goto idwINS01	; idwIORR		23
	goto idwINS01	; idwMOVR		24
	goto idwINS01	; idwROTL		25
	goto idwINS01	; idwROTR		26
	goto idwINS01	; idwSUBR		27
	goto idwINS01	; idwXORR		28
	goto idwINS01	; idwEXCHR		29
	goto idwINS01	; idwCOMPR		30
	goto wiEndExec	; NO INS		31

	org 0x2100

DecodeWIF2	movlw HIGH DecodeWIF2
	movwf PCLATH
	movf wifTblMain,W
	andlw 0x7F
	addwf PCL
	goto idwINS10	; idwGOSUB		32
	goto idwINS10	; idwGOTO		33
	goto idwINS10	; idwRETURN		34
	goto idwINS10	; idwRETVAL		35
	goto idwINS10	; idwHALT		36
	goto idwINS10	; idwSTSKIP		37
	goto idwINS10	; idwSTFLAG		38
	goto idwINS10	; idwSETB		39
	goto idwINS10	; idwCLRB		40
	goto idwINS10	; idwTOGB		41
	goto idwINS10	; idwSKIPBS		42
	goto idwINS10	; idwSKIPBC		43
	goto idwINS10	; idwROMaddr	44
	goto wiEndExec	; NO INS		45
	goto wiEndExec	; NO INS		46
	goto wiEndExec	; NO INS		47
	goto idwSKIPIF	; idwSKIPIF		48
	goto idwREADID	; idwREADID		49
	goto idwRESETI	; idwRESETI		50
	goto idwRDRSTI	; idwRDRSTI		51
	goto idwSKIPOF	; idwSKIPOF		52
	goto idwWRITOD	; idwWRITOD		53
	goto idwRESETO	; idwRESETO		54
	goto idwWTRSTO	; idwWTRSTO		55
	goto idwSEROUT	; idwSEROUT		56
	goto idwSKIPNS	; idwSKIPNS		57
	goto idwSERACC	; idwSERACC		58
	goto idwROMACC	; idwROMACC		59
	goto idwSKIPTS	; idwSKIPTS		60
	goto idwTIMER	; idwTIMER		61
	goto idwEEPROM	; idwEEPROM		62
;	goto wiEndExec	; NO INS		63

	org 0x2200

idwINS00	movlw HIGH idwINS00
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwINS001
	goto idwINS002
	goto idwINS003
	goto idwINS004

idwINS003	bsf rfdeLEDs,winWrite
	movlw HIGH idwINS003
	movwf PCLATH
	bcf rfdeLEDs,winExec
	movf wifTblMain,W
	addwf PCL
	goto idwNOP3
	goto idwADDVA3
	goto idwANDVA3
	goto idwIORVA3
	goto idwMOVVA3
	goto idwSUBVA3
	goto idwXORVA3
	goto idwSKIPA3
	goto idwRDSWA3
	goto idwDAAC3
	goto idwMULTA3
	goto idwRANDA3
	goto wiEndExec		; no ins
	goto wiEndExec		; no ins
	goto wiEndExec		; no ins
	goto wiEndExec		; no ins

idwINS004	movlw HIGH idwINS004
	movwf PCLATH
	movf wifTblMain,W
	addwf PCL
	goto incSkipPC
	goto idwMBAC4
	goto idwMBAC4
	goto idwMBAC4
	goto idwMBAC4
	goto idwMBAC4
	goto idwMBAC4
	goto incSkipPC
	goto idwMBAC4
	goto idwMBAC4
	goto idwMULTA4
	goto idwRANDA4
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC

idwINS01	movlw HIGH idwINS01
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwINS011
	goto idwINS012
	goto idwINS013
	goto idwINS014

	org 0x2300

idwINS013	bsf rfdeLEDs,winWrite
	movlw HIGH idwINS013
	movwf PCLATH
	bcf rfdeLEDs,winExec
	movf wifTbl013,W
	addwf PCL
	goto idwADDR3
	goto idwANDR3
	goto idwCLRR3
	goto idwDECR3
	goto idwDECSZ3
	goto idwINCR3
	goto idwINCSZ3
	goto idwIORR3
	goto idwMOVR3
	goto idwROTL3
	goto idwROTR3
	goto idwSUBR3
	goto idwXORR3
	goto idwEXCHR3
	goto idwCOMPR3
	goto wiEndExec		; no ins

idwINS10	movlw HIGH idwINS10
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwINS101
	goto idwINS102
	goto idwINS103
	goto idwINS104

idwINS102	bcf rfdeLEDs,winRead
	movlw HIGH idwINS102
	movwf PCLATH
	bsf rfdeLEDs,winExec
	movf wifTbl1034,W
	addwf PCL
	goto idwGOSTO2
	goto idwGOSTO2
	goto idwNxCyc10
	goto idwNxCyc10
	goto idwNxCyc10
	goto idwNxCyc10
	goto idwNxCyc10
	goto idwSCSK2
	goto idwSCSK2
	goto idwSCSK2
	goto idwSCSK2
	goto idwSCSK2
	goto idwROMADD2
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC

	org 0x2400

idwINS103	bsf rfdeLEDs,winWrite
	movlw HIGH idwINS103
	movwf PCLATH
	bcf rfdeLEDs,winExec
	movf wifTbl1034,W
	addwf PCL
	goto idwGOSUBb
	goto idwNxCyc10		; GOTO
	goto idwRETURNb
	goto idwRETURNb
	goto idwNxCyc10		; Halt
	goto idwSTSKIP
	goto idwSTFLAG
	goto idwSETBb
	goto idwCLRBb
	goto idwTOGBb
	goto idwSKIPBSb
	goto idwSKIPBCb
	goto idwROMaddr
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC

idwINS104	movlw HIGH idwINS104
	movwf PCLATH
	movf wifTbl1034,W
	addwf PCL
	goto finGGRR
	goto finGGRR
	goto finGGRR
	goto idwRETVALc
	goto idwHALTc
	goto incSkipPC
	goto incSkipPC
	goto srcTOtgt		; SETB (CLRB TOGB) need SRC set as TGT
	goto srcTOtgt		; CLRB
	goto srcTOtgt		; TOGB
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC
	goto incSkipPC		; no ins
;
; --------------------------------
; SUBROUTINE: SET CARRY ZERO FLAGS
; --------------------------------
;
doCZflags	bcf statusWIF,wifCarry
	btfsc STATUS,C
	bsf statusWIF,wifCarry
doZflag	bcf statusWIF,wifZero
	btfsc STATUS,Z
	bsf statusWIF,wifZero
	return
;
; -------------------------
; SUBROUTINE: SET ZERO FLAG
; -------------------------
;
doCflag	bcf statusWIF,wifCarry
	btfsc STATUS,C
	bsf statusWIF,wifCarry
	return
;
; inc PC with skip
;
incSkipPC	btfsc genBitsWIF,wifPCmod	; did PC get modified with [P] [TGT]
	goto wiEndExec		; yes

	btfsc genBits1,fgNoPCPLusOne	; do increment PC after instruction completes ??
	goto pcSKIP			; no, just do skip (Eg ROMACC instruction)

doSkipPC	incfsz PCregL
	goto pcSKIP

	incf PCregH,W
	andlw wiRomHI
	movwf PCregH
pcSKIP	btfss genBitsWIF,wifSkip
	goto wiEndExec

	bcf genBitsWIF,wifSkip
	goto doSkipPC		; skip set, increment PC (again)
;
; instruction decode TYPE 00
;
idwINS001	bsf rfdeLEDs,winRead		; common to all Type 00 instructions
	movf wifADDR,W
	xorlw picNOP
	movlw insVAL		; VAL type instruction
	btfsc STATUS,Z
	movlw insGEN		; NOP is a GEN type instruction
	iorwf insLEDs

idwNxCyc00	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwINS00

idwINS002	bsf rfdeLEDs,winExec		; common to all Type 00 instructions
	movf wifADDR,W
	bcf rfdeLEDs,winRead
	xorlw picDAAC
	btfsc STATUS,Z
	goto idwNxCyc00

	movf wiCodeL,W
	movwf MAreg
	goto idwNxCyc00

; Type 00 cycle 3
idwNOP3	goto idwNxCyc00

idwADDVA3	bcf statusWIF,wifDC
	movf ACreg,W
	addwf MAreg,W
	btfsc STATUS,DC
	bsf statusWIF,wifDC 
	movf ACreg,W
	addwf MAreg,W
	movwf MBreg
	call doCZflags
	goto idwNxCyc00

idwANDVA3	movf MAreg,W
	andwf ACreg,W
	movwf MBreg
	call doZflag
	goto idwNxCyc00

idwIORVA3	movf MAreg,W
	iorwf ACreg,W
	movwf MBreg
	call doZflag
	goto idwNxCyc00

idwMOVVA3	movf MAreg,W
	movwf MBreg
	goto idwNxCyc00

idwSUBVA3	bcf statusWIF,wifDC 
	movf MAreg,W
	subwf ACreg,W
	btfsc STATUS,DC
	bsf statusWIF,wifDC 
	movwf MBreg
	call doCZflags
	goto idwNxCyc00

idwXORVA3	movf MAreg,W
	xorwf ACreg,W
	movwf MBreg
	call doZflag
	goto idwNxCyc00

idwSKIPA3	movf wiCodeH,W
	andlw 0x03
	btfss STATUS,Z
	goto SKgrlteq		; SKIPA functions <, =, >=

	btfss wiCodeL,0		; wiCodeL[0]=1 POS, wiCodeL[1]=1 NEG
	goto tstSKneg

	btfss ACreg,7		; POS
	bsf genBitsWIF,wifSkip
	goto idwNxCyc00

tstSKneg	btfss wiCodeL,1
	goto idwNxCyc00

	btfsc ACreg,7		; NEG
	bsf genBitsWIF,wifSkip
	goto idwNxCyc00

SKgrlteq	bsf genBitsWIF,wifSkip		; preset function succeeds
	btfsc wiCodeH,1
	goto SKlteq

	movf wiCodeL,W		; AC >= Val?
	subwf ACreg,W
	btfss STATUS,C
	bcf genBitsWIF,wifSkip		; no
	goto idwNxCyc00

SKlteq	btfss wiCodeL,0
	goto SKeq

	movf wiCodeL,W		; AC < Val?
	cpfslt ACreg
	bcf genBitsWIF,wifSkip		; no
	goto idwNxCyc00

SKeq	movf ACreg,W		; AC = Val?
	xorwf wiCodeL,W
	btfss STATUS,Z
	bcf genBitsWIF,wifSkip		; no
	goto idwNxCyc00

idwRDSWA3	movf binSwitch,W
	movwf MBreg
	bcf statusWIF,wifZero
	btfsc STATUS,Z
	bsf statusWIF,wifZero
	goto idwNxCyc00

idwDAAC3	movf ACreg,W
	daw
	movwf MBreg
	bcf statusWIF,wifCarry
	btfsc STATUS,C
	bsf statusWIF,wifCarry
	goto idwNxCyc00

idwMULTA3	movf wiCodeH,W
	andlw 0x03
	btfss STATUS,Z
	goto multaHL3

	movf MAreg,W		; MULTA [VAL]
	mulwf ACreg
	movf PRODH,W
	movwf MAreg
	movf PRODL,W
	movwf MBreg
	goto idwNxCyc00

multaHL3	xorlw 0x01
	btfss STATUS,Z
	goto multaH3

	movf MULHreg,W		; MULTA [LA]
	movwf MBreg
	goto idwNxCyc00

multaH3	movf MULLreg,W		; MULTA [LA]
	movwf MBreg
	goto idwNxCyc00

idwRANDA3	clrf MBreg
	goto idwNxCyc00

; Type 00 cycle 4
idwMBAC4	movf MBreg,W
	movwf ACreg
	goto incSkipPC

idwMULTA4	movf wiCodeH,W
	andlw 0x03
	btfss STATUS,Z
	goto multaHL4

	movf MAreg,W		; MULTA [VAL]
	movwf MULHreg
	movf MBreg,W
	movwf MULLreg
	goto incSkipPC

multaHL4	movf MBreg,W		; MULTA [HA] [LA]
	movwf ACreg
	goto incSkipPC

idwRANDA4	movf TMR0L,W		; psuedo random number
	mulwf TMR1L
	movf PRODL,W
	movwf ACreg
	bcf statusWIF,wifZero
	btfsc STATUS,Z
	bsf statusWIF,wifZero
	goto incSkipPC
;
; instruction decode TYPE 01
;
idwINS011	bsf rfdeLEDs,winRead		; common to all Type 01 instructions
	bsf insLEDs,winREG		; 0000 00ss ttxr rrrr
idwNxCyc01	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwINS01

idwINS012	bsf rfdeLEDs,winExec		; common to all Type 00 instructions
	call GetSRCreg
	movwf MAreg
	bcf rfdeLEDs,winRead
	call GetTGTreg
	movwf MBreg
	goto idwNxCyc01

idwINS014	call codeRslt		; see if instruction has a result to MEM bit
 	btfss genBits2,n_ack
	goto do014TGT

	btfss wiCodeL,5		; if instruction bit 5 = 0 then result to SRC
	call matTGTtoSRC

do014TGT	call PutTGTreg		; common to all Type 00 instructions
	goto incSkipPC

idwADDR3	bcf statusWIF,wifDC
	movf MAreg,W
	addwf MBreg
	btfsc STATUS,DC
	bsf statusWIF,wifDC 
	call doCZflags
	goto idwNxCyc01

idwANDR3	movf MAreg,W
	andwf MBreg
	call doZflag
	goto idwNxCyc01

idwCLRR3	clrf MBreg
	bsf statusWIF,wifZero
	goto idwNxCyc01

idwDECR3	decf MAreg,W
	movwf MBreg
	call doZflag
	goto idwNxCyc01

idwDECSZ3	decf MAreg,W
	movwf MBreg
	btfsc STATUS,Z
	bsf genBitsWIF,wifSkip
	goto idwNxCyc01

idwINCR3	incf MAreg,W
	movwf MBreg
	call doZflag
	goto idwNxCyc01

idwINCSZ3	incf MAreg,W
	movwf MBreg
	btfsc STATUS,Z
	bsf genBitsWIF,wifSkip
	goto idwNxCyc01

idwIORR3	movf MAreg,W
	iorwf MBreg
	call doZflag
	goto idwNxCyc01

idwMOVR3	movf MAreg,W
	movwf MBreg
	call doZflag
	goto idwNxCyc01

idwROTL3	bcf STATUS,C
	btfsc statusWIF,wifCarry
	bsf STATUS,C
	rlcf MAreg,W
	movwf MBreg
	bcf statusWIF,wifCarry
	btfsc STATUS,C
	bsf statusWIF,wifCarry
	goto idwNxCyc01

idwROTR3	bcf STATUS,C
	btfsc statusWIF,wifCarry
	bsf STATUS,C
	rrcf MAreg,W
	movwf MBreg
	bcf statusWIF,wifCarry
	btfsc STATUS,C
	bsf statusWIF,wifCarry
	goto idwNxCyc01

idwSUBR3	bcf statusWIF,wifDC 
	movf MAreg,W
	subwf MBreg
	btfsc STATUS,DC
	bsf statusWIF,wifDC 
	call doCZflags
	goto idwNxCyc01

idwXORR3	movf MAreg,W
	xorwf MBreg
	call doZflag
	goto idwNxCyc01

idwEXCHR3	swapf MAreg,W
	movwf MBreg
	goto idwNxCyc01

idwCOMPR3	movf MAreg,W
	xorlw 0xFF
	movwf MBreg
	call doZflag
	goto idwNxCyc01
;
; instruction decode TYPE 10
;
idwINS101	bsf rfdeLEDs,winRead		; common to all Type 01 instructions
	bsf insLEDs,winGEN		; 0000 00ss ttxr rrrr
idwNxCyc10	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwINS10

idwGOSTO2	movf wiCodeL,W		; LO address
	movwf MAreg
	movf wiCodeH,W		; HI address
	andlw 0x03
	movwf MBreg
	goto idwNxCyc10

idwSCSK2	call GetSRCreg
	movwf MAreg
	call wiBitMask
            movwf MBreg
	goto idwNxCyc10

idwROMADD2	clrf MAreg			; if MaskROMADRHL = 1, MA = 1
	btfsc wiCodeH,0
	incf MAreg
	movf wiCodeL,W
	movwf MBreg
	goto idwNxCyc10

idwGOSUBb	movf stack1H,W
	movwf stack2H
	movf stack1L,W
	movwf stack2L

	movf stack0H,W
	movwf stack1H
	movf stack0L,W
	movwf stack1L

	movf PCregL,W
	movwf stack0L
	movf PCregH,W
	movwf stack0H

	infsnz stack0L
	incf stack0H
	movlw 0x03
	andwf stack0H
	goto idwNxCyc10

idwRETURNb	movf stack0H,W
	movwf MBreg
	movf stack0L,W
	movwf MAreg

	movf stack1H,W
	movwf stack0H
	movf stack1L,W
	movwf stack0L

	movf stack2H,W
	movwf stack1H
	movf stack2L,W
	movwf stack1L
	goto idwNxCyc10

idwSTSKIP	btfsc wiCodeL,0
	goto tstSTset

	btfss wiCodeL,1
	goto BSclrDC

	btfss statusWIF,wifCarry
	bsf genBitsWIF,wifSkip
	goto setPCMAB

BSclrDC	btfss wiCodeL,2
	goto BSclrZR

	btfss statusWIF,wifDC
	bsf genBitsWIF,wifSkip
	goto setPCMAB

BSclrZR	btfss statusWIF,wifZero
	bsf genBitsWIF,wifSkip
	goto setPCMAB

tstSTset	btfss wiCodeL,1
	goto BSsetDC

	btfsc statusWIF,wifCarry
	bsf genBitsWIF,wifSkip
	goto setPCMAB

BSsetDC	btfss wiCodeL,2
	goto BSsetZR

	btfsc statusWIF,wifDC
	bsf genBitsWIF,wifSkip
	goto setPCMAB

BSsetZR	btfsc statusWIF,wifZero
	bsf genBitsWIF,wifSkip

setPCMAB	movf PCregL,W
	movwf MAreg
	movf PCregH,W
	andlw 0x03
	movwf MBreg
	goto idwNxCyc10

idwSTFLAG	btfsc wiCodeL,0
	goto setStat10

	btfsc wiCodeL,1
	bcf statusWIF,wifCarry
	btfsc wiCodeL,2
	bcf statusWIF,wifDC
	btfsc wiCodeL,3
	bcf statusWIF,wifZero
	btfsc wiCodeL,4
	bcf statusWIF,wifRP
	goto idwNxCyc10

setStat10	btfsc wiCodeL,1
	bsf statusWIF,wifCarry
	btfsc wiCodeL,2
	bsf statusWIF,wifDC
	btfsc wiCodeL,3
	bsf statusWIF,wifZero
	btfsc wiCodeL,4
	bsf statusWIF,wifRP
	goto idwNxCyc10

idwSETBb	movf MAreg,W
	iorwf MBreg
	goto idwNxCyc10

idwCLRBb	movf MBreg,W
	xorlw 0xFF
	andwf MAreg,W
	movwf MBreg
	goto idwNxCyc10

idwTOGBb	movf MAreg,W
	xorwf MBreg
	goto idwNxCyc10

idwSKIPBSb	movf MAreg,W
	andwf MBreg,W
	btfss STATUS,Z
	bsf genBitsWIF,wifSkip
	goto idwNxCyc10

idwSKIPBCb	movf MAreg,W
	andwf MBreg,W
	btfsc STATUS,Z
	bsf genBitsWIF,wifSkip
	goto idwNxCyc10

idwROMaddr	btfsc wiCodeH,0
	goto HLbitHIGH

	movf MBreg,W		; transferring low ROMaddr byte
	movwf ROMaddrL
	goto idwNxCyc10

HLbitHIGH	movf MBreg,W		; transferring high ROMaddr byte
	andlw 0x03
	movwf ROMaddrH
	goto idwNxCyc10

srcTOtgt	call matTGTtoSRC		; SETB CLRB need TGT automatically matched to SRC
	call PutTGTreg
	goto incSkipPC
;
; instruction decode TYPE 10 - 4
;
idwRETVALc	movf wiCodeL,W
	movwf ACreg
finGGRR	movf MBreg,W
	movwf PCregH
	movf MAreg,W
	movwf PCregL
	goto wiEndExec

idwHALTc	bsf genBits1,fgHALT
	goto incSkipPC

	org 0x2900

; ---------------------------------------------------------------------
idwSKIPIF	movlw HIGH idwSKIPIF
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwSKIPIF1
	goto idwSKIPIF2
	goto idwSKIPIF3
	goto idwSKIPIF4
	goto incSkipPC

idwREADID	movlw HIGH idwREADID
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwREADID1
	goto idwREADID2
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwREADID3
	goto incSkipPC

idwRESETI	movlw HIGH idwRESETI
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwRESETI1
	goto idwRESETI2
	goto idwRESETI3
	goto idwRESETI4

idwSKIPOF	movlw HIGH idwSKIPOF
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwSKIPOF1
	goto idwSKIPOF2
	goto idwSKIPOF3
	goto idwSKIPOF4
	goto incSkipPC

idwRDRSTI	movlw HIGH idwRDRSTI
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwREADID1
	goto idwREADID2
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwRDID207
	goto idwREADID3
	goto idwRDRSTI1
	goto idwRDRSTI2
	goto idwRDRSTI3
	goto incSkipPC

idwRESETO	movlw HIGH idwRESETO
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwRESETO1
	goto idwRESETO2
	goto idwRESETO3
	goto idwRESETO4

	org 0x2A00

idwWRITOD	movlw HIGH idwWRITOD
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwWRITOD1
	goto idwWRITOD2
	goto idwWRITOD3
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto incSkipPC

idwWTRSTO	movlw HIGH idwWTRSTO
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwWTRSTI
	goto idwWTRST2
	goto idwWRITOD3
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTOD407
	goto idwWTRSTI1
	goto idwWTRSTI2
	goto idwWTRSTI3
	goto incSkipPC

idwSEROUT	movlw HIGH idwSEROUT
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwSEROUT1
	goto idwSEROUT2
	goto idwSEROUT3
	goto idwSEROUT4

idwSKIPNS	movlw HIGH idwSKIPNS
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwSKIPNS1
	goto idwSKIPNS2
	goto idwSKIPNS3
	goto idwSKIPNS4

idwSERACC	movlw HIGH idwSERACC
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwSERACC1
	goto idwSERACC2
	goto idwSERACC3
	goto idwSERACC4

idwROMACC	movlw HIGH idwROMACC
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwROMACC1
	goto idwROMACC2
	goto idwROMACC3
	goto idwROMACC4

	org 0x2B00

idwSKIPTS	movlw HIGH idwSKIPTS
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwSKIPTS1
	goto idwSKIPTS2
	goto idwSKIPTS3
	goto idwSKIPTS4

idwTIMER	movlw HIGH idwTIMER
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwTIMER1
	goto idwTIMER2
	goto idwTIMER3
	goto idwTIMER4

idwEEPROM	movlw HIGH idwEEPROM
	movwf PCLATH
	movf execIndex,W
	addlw 4h
	movwf execIndex
	addwf PCL
	goto idwEEP1
	goto idwEEP2
	goto idwEEP3
	goto idwEEP4
; ---------------------------------------------------------------------
idwSKIPIF1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwSKIPIF2	bsf rfdeLEDs,winExec
	call wifSetIOin
	bcf rfdeLEDs,winRead
	goto MainLoop

idwSKIPIF3	btfsc genBitsWIF,wifDev
	goto wiSKIPI1

	skipIN0RIhi			; skip IN FG dev 0
	bsf genBitsWIF,wifSkip
	goto MainLoop

wiSKIPI1
	skipIN1RIhi			; skip IN FG dev 1
	bsf genBitsWIF,wifSkip
	goto MainLoop

idwSKIPIF4	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec
	goto MainLoop

; ---------------------------------------------------------------------
idwREADID1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwREADID2	bsf rfdeLEDs,winExec
	call wifSetIOin
	bcf rfdeLEDs,winRead
	goto MainLoop

idwRDID207
	clrIOclk			; CLOCK = LO
	call CKdelay		; wait for stabilize

	btfsc genBitsWIF,wifDev
	goto wifDevIN1

	bcf STATUS,C
	skipIN0DAlo			; read DEV 0 INPUT
	bsf STATUS,C
	goto wifSHin

wifDevIN1	bcf STATUS,C
	skipIN1DAlo			; read DEV 1 INPUT
	bsf STATUS,C
wifSHin	rrcf MBreg
	setIOclk			; CLOCK = HI
	goto MainLoop

idwREADID3	bsf rfdeLEDs,winWrite
	movf MBreg,W
	bcf statusWIF,wifZero
	btfsc STATUS,Z
	bsf statusWIF,wifZero
	bcf rfdeLEDs,winExec
	call PutTGTreg
	goto MainLoop

; ---------------------------------------------------------------------
idwRDRSTI1	call wifSetIOin		; select input
	goto MainLoop

idwRDRSTI2
	clrIOrst			; DEV X RESET = LO
	goto MainLoop

idwRDRSTI3
	setIOrst			; DEV X RESET = HI
	goto incSkipPC

; ---------------------------------------------------------------------
idwRESETI1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwRESETI2	bsf rfdeLEDs,winExec
	call wifSetIOin		; select input
	bcf rfdeLEDs,winRead
	goto MainLoop

idwRESETI3	bsf rfdeLEDs,winWrite
	clrIOrst			; DEV X RESET = LO
	bcf rfdeLEDs,winExec
	goto MainLoop

idwRESETI4
	setIOrst			; DEV X RESET = HI
	goto incSkipPC

; ---------------------------------------------------------------------
idwSKIPOF1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwSKIPOF2	bsf rfdeLEDs,winExec
	call wifSetIOout
	bcf rfdeLEDs,winRead
	goto MainLoop

idwSKIPOF3	btfsc genBitsWIF,wifDev
	goto wiSKIPOH

	skipOUT0RIhi		; skip OUT FG dev 0 = HI
	bsf genBitsWIF,wifSkip
	goto MainLoop

wiSKIPOH
	skipOUT1RIhi		; skip OUT FG dev 1 = HI
	bsf genBitsWIF,wifSkip
	goto MainLoop

idwSKIPOF4	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec
	goto MainLoop

; ---------------------------------------------------------------------
idwRESETO1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	call wifSetIOout		; select output
	btfsc rfdeLEDs,ResetT1T13	; ResetT1T13 = 1, Reset on BO
	clrIOrst			; yes, DEV X RESET = LO
	goto MainLoop

idwRESETO2	bsf rfdeLEDs,winExec
	setIOrst			; DEV X RESET = HI, may be HI already but no matter
	bcf rfdeLEDs,winRead
	goto MainLoop

idwRESETO3	bsf rfdeLEDs,winWrite
	btfss rfdeLEDs,ResetT1T13	; ResetT1T13 = 0, Reset on AO
	clrIOrst			; yes, DEV X RESET = LO
	bcf rfdeLEDs,winExec
	goto MainLoop

idwRESETO4
	setIOrst			; DEV X RESET = HI, may be HI already but no matter
	goto incSkipPC

; ---------------------------------------------------------------------
idwWRITOD1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwWRITOD2	bsf rfdeLEDs,winExec
	call wifSetIOout		; select output
	bcf rfdeLEDs,winRead
	goto MainLoop

idwWRITOD3	bsf rfdeLEDs,winWrite
	call GetSRCreg
	movwf MBreg
	bcf rfdeLEDs,winExec
	goto MainLoop

idwWTOD407	bcf STATUS,C
	rrcf MBreg
	btfsc genBitsWIF,wifDev
	goto dev1shO

	btfsc STATUS,C
	goto idwDev0hi

	clrDataOut0			; dev 0 out = LO
	goto idwExeWTO

idwDev0hi	setDataOut0			; dev 0 out = HI
	goto idwExeWTO
	
dev1shO	btfsc STATUS,C
	goto idwDev1hi

	clrDataOut1			; dev 1 out = LO
	goto idwExeWTO

idwDev1hi	setDataOut1			; dev 1 out = HI

idwExeWTO	call CKdelay
	clrIOclk
	call CKdelay
	setIOclk
	goto MainLoop

; ---------------------------------------------------------------------
idwWTRSTI	bsf rfdeLEDs,winRead
	call wifSetIOout		; select output
	bsf insLEDs,winIO
	btfsc rfdeLEDs,ResetT1T13	; ResetT1T13 = 1, Reset on BO
	clrIOrst			; yes, DEV X RESET = LO
	goto MainLoop

idwWTRST2	bsf rfdeLEDs,winExec
	setIOrst			; DEV X RESET = HI, may be HI already but no matter
	bcf rfdeLEDs,winRead
	goto MainLoop

; ............

idwWTRSTI1	call wifSetIOout		; select output
	goto MainLoop

idwWTRSTI2	btfss rfdeLEDs,ResetT1T13	; ResetT1T13 = 0, Reset on AO
	clrIOrst			; yes, DEV X RESET = LO
	goto MainLoop

idwWTRSTI3
	setIOrst			; DEV X RESET = HI, may be HI already but no matter
	goto incSkipPC

; ---------------------------------------------------------------------
idwSEROUT1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwSEROUT2	bsf rfdeLEDs,winExec
	btfsc genBitsWIF,wifSerial
	call Rst232			; reset serial port if errors
	bcf rfdeLEDs,winRead
	goto MainLoop

idwSEROUT3	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec
	goto MainLoop

idwSEROUT4	btfss genBitsWIF,wifSerial
	goto incSkipPC

	movf ACreg,W
	call TransWt
	goto incSkipPC

; ---------------------------------------------------------------------
idwSKIPNS1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
	goto MainLoop

idwSKIPNS2	bsf rfdeLEDs,winExec
	nop
	bcf rfdeLEDs,winRead
	bcf genBitsWIF,wifSerRcvd
	goto MainLoop

idwSKIPNS3	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec

	btfss genBitsWIF,wifSerial
	goto MainLoop

	BANKSEL PIR3		; .receive
	btfss PIR3,RC1IF,BANKED		; (5) check for received data
	goto incSkipPC		; ignore SKIPNS4

	bsf genBitsWIF,wifSkip
	bsf genBitsWIF,wifSerRcvd
	movf RC1REG,W
	movwf MBreg
	btfsc STATUS,Z
	bsf statusWIF,wifZero
	goto MainLoop

idwSKIPNS4	btfss genBitsWIF,wifSerRcvd
	goto incSkipPC

	movf MBreg,W
	movwf ACreg
	goto incSkipPC

; ---------------------------------------------------------------------
idwSERACC1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
idwNxSACC	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwSERACC

idwSERACC2	bsf rfdeLEDs,winExec
	nop
	bcf rfdeLEDs,winRead
	goto idwNxSACC

idwSERACC3	bsf rfdeLEDs,winWrite
	call Rst232			; reset serial port if errors
	bcf rfdeLEDs,winExec
	goto idwNxSACC

idwSERACC4	bcf genBitsWIF,wifSerial
	movf wiCodeL,W
	andlw 0x01
	btfss STATUS,Z
	bsf genBitsWIF,wifSerial
	goto incSkipPC

; ---------------------------------------------------------------------
idwROMACC1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
idwNxRACC	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwROMACC

idwROMACC2	bsf rfdeLEDs,winExec
	nop
	bcf rfdeLEDs,winRead
	goto idwNxRACC

idwROMACC3	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec
	goto idwNxRACC

idwROMACC4	btfsc wiCodeL,2		; PI?
	call IncPtr			; yes, pre increment pointer
	btfsc wiCodeL,4		; MI?
	call DecPtr			; yes, pre decrement pointer

	movf ROMaddrH,W		; memory address from ROMaddr * 2 (2 bytes per memory)
	andlw 0x03
	movwf TempA
	movf ROMaddrL,W
	movwf TempB

	bcf STATUS,C
	rlcf TempB,W
	movwf FSR0L
	rlcf TempA,W
	addlw MEMbuff
	movwf FSR0H			; INDF points to memory HIGH byte

	btfsc wiCodeL,1		; W = 0, R = 1
	goto ReadRAcc

; write ROMACC
	btfsc wiCodeL,0		; L = 0, H = 1
	goto wtRAChi

	movf POSTINC0,W		; dummy read, INDF0 now points to memory LOW byte
	movf ACreg,W		; AC -> low byte
	movwf INDF0
	goto doRACinc

wtRAChi	movf ACreg,W		; AC -> high byte
	andlw 0x03
	movwf INDF0
	goto doRACinc

; read ROMACC
ReadRAcc	btfsc wiCodeL,0		; L = 0, H = 1
	goto rdRAChi

	movf POSTINC0,W		; dummy read, INDF0 now points to memory LOW byte
	movf INDF0,W		; low byte -> AC
	movwf ACreg
	goto doRACinc

rdRAChi	movf INDF0,W		; high byte -> AC
	movwf ACreg

doRACinc	btfsc wiCodeL,3		; IP?
	call IncPtr			; yes, post increment pointer
	btfsc wiCodeL,5		; IM?
	call DecPtr			; yes, post decrement pointer
	goto incSkipPC
;
; Increment ROMPTR
;
IncPtr	infsnz ROMaddrL
	incf ROMaddrH
	movlw 0x03
	andwf ROMaddrH
	return
;
; Decrement ROMPTR
;
DecPtr	decf ROMaddrL
	movlw 0xFF
	xorwf ROMaddrL,W
	btfss STATUS,Z
	return

	decf ROMaddrH
	movlw 0x03
	andwf ROMaddrH
	return

; ---------------------------------------------------------------------
idwSKIPTS1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
idwNxRSTS	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwSKIPTS

idwSKIPTS2	bsf rfdeLEDs,winExec
	nop
	bcf rfdeLEDs,winRead
	goto idwNxRSTS

idwSKIPTS3	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec

	btfss wiCodeL,0		; code[0] -> 1 = DEP sw, 0 = EXAM sw
	goto SKTSEXsw

; DEP SW  TRIG1 ------------------
	btfsc wiCodeL,1		; code[1] -> 0 = test if switch down, 1 = test if switch is up
	goto DPswUP

	btfsc ctrSWuse,ctrsDP		; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	bsf genBitsWIF,wifSkip		; NEEDS ANDing WITH EXTERNAL TRIG1 SWITCH TOO <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	goto incSkipPC
				; need new PCB to finish switch read code which will AND these values
DPswUP	btfss ctrSWuse,ctrsDP
	bsf genBitsWIF,wifSkip		; NEEDS ANDing WITH EXTERNAL TRIG1 SWITCH TOO <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
	goto incSkipPC		; ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

; EXAM SW  TRIG2 ------------------
SKTSEXsw	btfsc wiCodeL,1		; code[1] -> 0 = test if switch down, 1 = test if switch is up
	goto EXswUP

	btfsc ctrSWuse,ctrsEX
	bsf genBitsWIF,wifSkip
	goto incSkipPC

EXswUP	btfss ctrSWuse,ctrsEX
	bsf genBitsWIF,wifSkip
	goto incSkipPC

idwSKIPTS4	goto incSkipPC

; ---------------------------------------------------------------------
idwTIMER1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
idwNxRSTT	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwTIMER

idwTIMER2	bsf rfdeLEDs,winExec
	nop
	bcf rfdeLEDs,winRead
	goto idwNxRSTT

idwTIMER3	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec
	btfsc wiCodeH,0		; 0=setDelay, 1=SKIPOF
	goto T3skipof

	movf wiCodeL,W
	movwf delTIMER
	movwf valTIMER
	bcf genBits2,fgTimer		; reset TIMER flag
	goto idwNxRSTT

T3skipof	btfsc genBits2,fgTimer
	bsf genBitsWIF,wifSkip
	bcf genBits2,fgTimer		; reset TIMER flag
	goto idwNxRSTT

idwTIMER4	goto incSkipPC

; ---------------------------------------------------------------------
idwEEP1	bsf rfdeLEDs,winRead
	bsf insLEDs,winIO
idwNxRSTE	btfss genBits1,fgSlowFast
	goto MainLoop
	goto idwEEPROM

idwEEP2	bsf rfdeLEDs,winExec
	nop
	bcf rfdeLEDs,winRead
	clrf MBreg
	btfss wiCodeH,0		; wait for eeprom to finish writing?
	goto chkEEadd

	clrf MAreg
	goto idwNxRSTE
	
chkEEadd	btfss wiCodeL,7		; EEPROM address?
	goto intEEad1		; no

	movf wiCodeL,W
	andlw wiEEPmax
	movwf MAreg
	goto idwNxRSTE

intEEad1	movf intEEPaddr,W
	andlw wiEEPmax
	movwf MAreg
	goto idwNxRSTE

idwEEP3	bsf rfdeLEDs,winWrite
	nop
	bcf rfdeLEDs,winExec
	btfss wiCodeL,6		; EEPROM read?
	goto intEEwt3		; no

	call wiEEPaddr
	call RDeeprom
	movwf MBreg
	goto idwNxRSTE

intEEwt3	btfss wiCodeL,5		; EEPROM write?
	goto idwNxRSTE		; no

	movf ACreg,W
	movwf MBreg
	goto idwNxRSTE

idwEEP4	btfss wiCodeL,7		; EEPROM address?
	goto intEEad4		; no

	movf MAreg,W
	movwf intEEPaddr
	goto incSkipPC

intEEad4	btfss wiCodeL,6		; EEPROM read?
	goto intEEwt4		; no

	movf MBreg,W
	btfsc STATUS,Z
	bsf statusWIF,wifZero
	movwf ACreg
	goto incEEadr

intEEwt4	btfss wiCodeL,5		; EEPROM write?
	goto EEPendWT

	btfsc NVMCON1,WR		; if eeprom is writing then no new write else write to eeprom
	goto incSkipPC

	call wiEEPaddr
	movf MBreg,W
	call WTeeprom
incEEadr	incf intEEPaddr,W
	andlw wiEEPmax
	movwf intEEPaddr
	goto incSkipPC

EEPendWT	btfss wiCodeH,0		; skip on EEPROM write complete?
	goto incSkipPC		; no

	btfss NVMCON1,WR		; if eeprom is writing then no skip
	bsf genBitsWIF,wifSkip
	goto incSkipPC

; ---------------------------------------------------------------------
wiEndExec	bcf genBitsWIF,wifNewINS
	bcf genBits1,fgNoPCPLusOne
	bcf genBitsWIF,wifPCmod
	btfsc genBits1,fgHALT		; Cycle 23 stop execution if HALT set
	goto doHALTwi

	btfss genBits1,fgSinCon		; stop execution if SINGLE set
	goto doHALTwi

	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetRun		; EXEC OFF, RUN STAYS ON
	goto FinExndWi

doHALTwi	bcf genBits1,fgHALT
	clrf insLEDs		; all instruction LEDs off
	movf rfdeLEDs,W
	andlw maskRFDEoff
	iorlw maskSetFetch		; EXEC OFF, RUN OFF, FTECH ON

FinExndWi	movwf rfdeLEDs
	movlw initCycleIndex
	movwf execIndex
	btfss genBits1,fgSlowFast
	clrf insLEDs		; all instruction LEDs off
	goto MainLoop
;
; -----------------------------------
; ENABLE TMR1 FOR "TIMER" INSTRUCTION
; -----------------------------------
;
TMR1SET	btfss rfdeLEDs,wiWIF
	goto offTIMER

	BANKSEL PMD1
	movlw b'11111100'		; tmr0 = ON, tmr1 = ON
	movwf PMD1,BANKED
	movf delTIMER,W		; set init delay - might be random
	btfsc STATUS,Z
	addlw d'50'			; if was zero = default 500mS delay
	movwf valTIMER
	bcf genBits2,fgTimer		; reset TIMER flag
	movlw b'00000001'		; enable TMR1 - Fosc/4
	movwf TMR1CLK
	movlw b'00010001'		; TMR1 ON PS 1:2
	movwf T1CON
	BANKSEL PIR4
	bcf PIR4,TMR1IF,BANKED
	return

offTIMER
	BANKSEL PMD1
	movlw b'11111110'		; tmr0 = ON, tmr1 = OFF
	movwf PMD1,BANKED
	return

	org 0x3000

; ---------------------------------------
pcTest	goto PCtest2
	movlw rzMainCmdTest
	; call TransWt

	movf binSwitch,W
	call TransWt
	goto MainLoop
;	movlw LOW IntEEPwif
;	movwf NVMADRL
;	movlw HIGH IntEEPwif
;	movwf NVMADRH
;	movlw 0x20
;	movwf TempA
;pctLoop	call RDeeprom
;	call TransWt
;	decfsz TempA
;	goto pctLoop
;	goto MainLoop
;	movlw RAMbuff
;	movwf FSR0H
;	movlw 0x00
;	movwf FSR0L
;	movlw 0x40
;	movwf TempA
;pctLoop	movf POSTINC0,W
;	call TransWt
;	decfsz TempA
;	goto pctLoop
;	goto MainLoop


;	clrf NVMADRH
;	movlw 0h
;	movwf NVMADRL
;	movlw 9h
;	movwf TempC
;pcTestLoop	call RDeeprom		; returns ZP byte
;	call TransWt
;	decfsz TempC
;	goto pcTestLoop
;	movf rfdeLEDs,W
;	call TransWt
;	goto MainLoop

	org 0x3FE8

#include "copyright.asm"

	org 0x4000

; **************************************
; ******** secondary boot loader *******
; **************************************
;
; ----------------------------------
; SUBROUTINE: BLANK THE DISPLAY LEDS
; ----------------------------------
;
BlankDisp2	clrf PORTD			; all LED anodes LOW
	return
;
; ----------------------------------------
; SUBROUTINE: RECEIVE CHARACTER FROM RS232 - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; ----------------------------------------
; This routine does not return until a character is received.
;
Receive2
	BANKSEL PIR3		; .receive
RXwait2	btfss PIR3,RC1IF,BANKED		; (5) check for received data
	goto RXwait2

	movf RC1REG,W
	movwf TempA
	return
;
; ------------------------------------------------
; SUBROUTINE: WAIT UNTIL RS232 IS FINISHED SENDING - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; ------------------------------------------------
;
TransWt2	movwf TX1REG		; .transwt
TXwait2	btfss TX1STA,TRMT		; (1) transmission is complete if hi
	goto TXwait2
	return
;
; --------------------------------
; SUBROUTINE: READ INTERNAL EEPROM - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; --------------------------------
; address/data set up in EE regs
RDeeprom2	movlw b'00000000'		; .rdeeprom Setup Data EEPROM Access, reading
	movwf NVMCON1
	bsf NVMCON1,RD		; Issue EE Read
	movf NVMDAT,W		; W = EE_DATA
	infsnz NVMADRL
	incf NVMADRH
	return 
;
; ---------------------------------
; SUBROUTINE: WRITE INTERNAL EEPROM - ACCESSED BY BOOT CODE TOO - DON'T MOVE
; ---------------------------------
; address/data set up in EE regs
WTeeprom2	movwf NVMDAT
	movlw b'00000100'		; .wteeprom Setup Data EEPROM Access, enable writes
	movwf NVMCON1
	
	movlw 0x55
	movwf NVMCON2
	movlw 0xAA
	movwf NVMCON2
	bsf NVMCON1,WR		; Set WR bit to begin write

WtIEEPwt2	btfsc NVMCON1,WR		; Wait for write to complete
	goto WtIEEPwt2

	infsnz NVMADRL
	incf NVMADRH
	return 
;
; ----------------------------------------------
; NEW BOOT CODE ROM AREA - FOR UPDATING BOOT ROM
; ----------------------------------------------
;
doBootCode2	call BlankDisp2
	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	clrf TBLPTRU		; this boot programming starts at 0x000000
	clrf TBLPTRH
	clrf TBLPTRL
	movlw rzMainCmdGoProg		; ACK prog ROM
TXloop2	call TransWt2
;
; ----------------------
; WAITING FOR PC COMMAND
; ----------------------
;
WaitCom2	movlw HIGH WaitCom2
	movwf PCLATH
	call Receive		; wait and read from RS232
	andlw 0x07
	mullw 4h			; PCL offset for GOTO
	movf PRODL,W
	addwf PCL
	goto WaitCom2		;			(0)
	goto WaitCom2		;			(1)
	goto FillBuff2		; fill buffer		(2)
	goto WriteProg2		; write the buffer data		(3)
	goto WaitCom2		;			(4)
	goto WaitCom2		;			(5)
	goto WaitCom2
	goto EndProg2		; finished with prog ROM	(7)
;
; -------------------
; FILL PROGRAM BUFFER
; -------------------
;
FillBuff2	movlw d'4'			; 4 bytes at a time - no Xtal for OSC - minimising data transfer size
	movwf TempB
FBloop2	call Receive2
	movwf POSTINC0
	decfsz TempB
	goto FBloop2
 
	movlw rzBootCmdFill
	goto TXloop2
;
; -----------------------
; PROGRAM BUFFER INTO ROM
; -----------------------
;
WriteProg2	SetReadROM			; test if ROM is same as program data, no program if so
	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	movf TBLPTRU,W
	movwf TempA
	movf TBLPTRH,W
	movwf TempB
	movf TBLPTRL,W
	movwf TempC
	movlw d'128'
	movwf TempD

verfLoop2	tblrd*+
	movf TABLAT,W
	xorwf POSTINC0,W
	btfss STATUS,Z
	goto notVerf2

	decfsz TempD
	goto verfLoop2

	call IncAddr
	movlw rzBootCmdNoProg		; ACK no prog change
	goto TXloop2
;
; now do programming
;
notVerf2	movf TempA,W
	movwf TBLPTRU
	movf TempB,W
	movwf TBLPTRH
	movf TempC,W
	movwf TBLPTRL

	movlw b'10010100'		; erase block
	movwf NVMCON1
	nop
	movlw 0x55
	movwf NVMCON2
	movlw 0xAA
	movwf NVMCON2
	bsf NVMCON1,WR		; start erase (CPU stalls)

	tblrd*-			; set back by 1 - ensures that upper bit still access proper write block
	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L
	movlw d'128'
	movwf TempD
btProgLoop2	movf POSTINC0,W		; get low byte of buffer data
	movwf TABLAT		; present data to table latch
	tblwt+*			; write data, perform a short write to internal TBLWT holding register.
	decfsz TempD		; loop until holding registers are full
	goto btProgLoop2

	movlw b'10000100'		; write block
	movwf NVMCON1
	nop
	movlw 0x55
	movwf NVMCON2
	movlw 0xAA
	movwf NVMCON2
	bsf NVMCON1,WR		; start write (CPU stalls)

	call IncAddr2
	movlw rzBootCmdProg		; ACK
	goto TXloop2
;
; ------------------------------
; INCREMENT ADDRESS BY 128 BYTES
; ------------------------------
;
IncAddr2	movlw XFERbuff		; base address for ROM data buffer storage
	movwf FSR0H
	clrf FSR0L

	movf TempA,W
	movwf TBLPTRU
	movf TempB,W
	movwf TBLPTRH
	movf TempC,W
	movwf TBLPTRL

	movlw 0x80
	addwf TBLPTRL
	btfss STATUS,C
	return

	infsnz TBLPTRH
	incf TBLPTRU
	return
;
; --------------------------------
; SUBROUTINE: ERASE PROGRAM MEMORY
; --------------------------------
;
EraseMem	call BlankDisp
	movlw MEMbuff
	movwf FSR0H
	clrf FSR0L

	clrf TempB			; fill all memory with zero (EDUC8 and WIF)
	movlw d'8'			; 8 x 256 bytes
	movwf TempC
FMVloop	clrf POSTINC0
	decfsz TempB
	goto FMVloop
	decfsz TempC
	goto FMVloop

	
	clrf PCregL
	clrf PCregH
	clrf MAreg
	clrf MBreg
	clrf ACreg
	return
;
; ---------------------------------------------------------------------
; SUBROUTINE: SET SOME REGISTERS TO A DEFAULT STATE IF NOT ERASE MEMORY
; ---------------------------------------------------------------------
;
DefErase	clrf PCregL
	clrf PCregH
	movlw 0xFF			; 0737
	movwf MAreg
	movlw b'00110000'		; 0120
	movwf MBreg
	movlw b'00110010'		; 0122
	movwf ACreg
	return
;
; -----------------------------------------
; SUBROUTINE: SETUP INTERNAL EEPROM ADDRESS
; -----------------------------------------
; EEPROM Address MAreg
wiEEPaddr	movf MAreg,W
	andlw wiEEPmax
	addlw LOW IntEEPwif
	movwf NVMADRL
	movlw HIGH IntEEPwif
	movwf NVMADRH
	return
;
; -----------------------
; FINISHED WITH BOOT CODE
; -----------------------
;
EndProg2	bcf NVMCON1,WREN
	goto Start			; do a restart in main code

; testcode
educFirst	movlw MEMbuff
	movwf FSR0H
	clrf FSR0L
	movlw 0xEC		; CLA.CMA
	movwf POSTINC0
	movlw 0xFC		; SZA.SMA
	movwf POSTINC0
	movlw 0x29		; TAD E011
	movwf POSTINC0
	movlw 0x0A		; AND E012
	movwf POSTINC0
	movlw 0xE2		; RAL
	movwf POSTINC0
	movlw 0x2B		; TAD E013
	movwf POSTINC0
	movlw 0xF8		; SZA
	movwf POSTINC0
	movlw 0xA0		; JMP E000
	movwf POSTINC0
	movlw 0xE1		; IAC
	movwf POSTINC0
	movlw 0xA0		; JMP E000
	movwf POSTINC0
	movlw 0x03		; AND E003
	movwf POSTINC0
	movlw 0xEF		; ???
	movwf POSTINC0
	movlw 0xFA		; ???
	movwf POSTINC0
	return


PCtest2	movlw rzMainCmdTest
	call TransWt

	;selIOin0

	call Receive
	movf TempA,W
	btfsc STATUS,Z
	goto CLKoff

	setDataOut1
	movlw 0h
	call TransWt
	goto MainLoop

CLKoff
	clrDataOut1
	movlw 0h
	call TransWt
	goto MainLoop


	org 0x310000

; EEPROM DATA
; 0 = USER FLAGS
; EEflgWIF    Normal WIF mode   - bit 0      - 1 = Nrm, 0 = Wif

; EEflgZPM    Zero Page Mode    - bit 2      - 1 = OFF, 0 = ON
; EEflgSED    Allow slow ex/dp  - bit 3      - 1 = NO,  0 = YES
;             Slow Speed        - bits 5 4   - 500mS
; EEflgT13    T1 T13 Reset      - bit 6      - 1 = T13, 0 = T1
; EEflgRst    Clear registers on reset       - 1 = OFF, 0 = ON

	de 0xEF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF	; Regs = Zero, ZP=OFF, default slow speed, Allow slow ex/dp = OFF, T13
	de 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF	; WIF 8 byte storage

	end